from __future__ import annotations
import asyncio
from dataclasses import dataclass
import logging
import unittest

from openpectus.aggregator.aggregator_message_handlers import AggregatorMessageHandlers
from openpectus.aggregator.data import database
from openpectus.aggregator.frontend_publisher import FrontendPublisher
from openpectus.aggregator.models import SystemStateEnum
from openpectus.engine.engine import Engine
from openpectus.engine.engine_message_handlers import EngineMessageHandlers
from openpectus.engine.engine_message_builder import EngineMessageBuilder
from openpectus.protocol.aggregator_dispatcher import AggregatorDispatcher
import openpectus.protocol.aggregator_messages as AM
from openpectus.protocol.engine_dispatcher import EngineDispatcher
from openpectus.engine.engine_runner import EngineRunner
import openpectus.protocol.engine_messages as EM
from openpectus.protocol.exceptions import ProtocolException, ProtocolNetworkException
import openpectus.protocol.messages as M
from fastapi_websocket_rpc.logger import get_logger
from openpectus.aggregator.aggregator import Aggregator

from openpectus.test.engine.test_engine import create_test_uod
import openpectus.aggregator.data.models as DMdl


logging.basicConfig()
logger = get_logger("Test")
logger.setLevel(logging.DEBUG)

logging.getLogger('asyncio').setLevel(logging.ERROR)
logging.getLogger('openpectus.engine.engine_message_handlers').setLevel(logging.DEBUG)
logging.getLogger('openpectus.protocol.engine_dispatcher').setLevel(logging.DEBUG)
logging.getLogger('openpectus.engine.engine_runner').setLevel(logging.DEBUG)
logging.getLogger('openpectus.aggregator.aggregator').setLevel(logging.DEBUG)
logging.getLogger('openpectus.protocol.aggregator_dispatcher').setLevel(logging.DEBUG)


class ProtocolIntegrationTestCase(unittest.IsolatedAsyncioTestCase):
    ctx: ProtocolIntegrationTestCase.Context | None = None

    @dataclass
    class Context:
        """ Test helper class """
        engine: Engine
        engineDispatcher: EngineTestDispatcher
        engineRunner: EngineRunner
        engineMessageHandlers: EngineMessageHandlers

        aggregatorDispatcher: AggregatorTestDispatcher
        frontendPublisher: FrontendPublisher
        aggregator: Aggregator

    async def asyncSetUp(self) -> None:
        assert self.ctx is None
        await super().asyncSetUp()

        # setup in-memory database
        database.configure_db("sqlite:///:memory:")
        DMdl.DBModel.metadata.create_all(database._engine)  # type: ignore

        # the test class IsolatedAsyncioTestCase provides the event loop
        loop = asyncio.get_running_loop()

        uod = create_test_uod()
        engine = Engine(uod)
        engineDispatcher = EngineTestDispatcher()
        engineRunner = EngineRunner(engineDispatcher, EngineMessageBuilder(engine), engine.emitter, loop)
        engineMessageHandlers = EngineMessageHandlers(engine, engineDispatcher)

        aggregatorDispatcher = AggregatorTestDispatcher()
        frontendPublisher = FrontendPublisher()
        aggregator = Aggregator(aggregatorDispatcher, frontendPublisher)
        _ = AggregatorMessageHandlers(aggregator)

        # connect the two test dispatchers for direct communication
        aggregatorDispatcher.engineDispatcher = engineDispatcher
        engineDispatcher.aggregatorDispatcher = aggregatorDispatcher

        self.ctx = ProtocolIntegrationTestCase.Context(
            engine, engineDispatcher, engineRunner, engineMessageHandlers,
            aggregatorDispatcher, frontendPublisher, aggregator)

        # not sure why this is necessary. asyncTeardown should be doing this already
        self.addCleanup(lambda : engine.stop())
        await self.start_engine()

    async def asyncTeardown(self) -> None:
        await self.stop()
        await super().asyncTearDown()

    async def start_engine(self):
        assert self.ctx is not None
        self.ctx.engine.run()

    async def stop(self):
        assert self.ctx is not None
        await self.ctx.engineDispatcher.disconnect_async()
        await self.ctx.engineRunner.shutdown()

    async def register(self) -> str:
        assert self.ctx is not None

        register_message = EM.RegisterEngineMsg(
            computer_name="test_host",
            uod_name="test_uod_name",
            uod_author_name="uod-author-name",
            uod_author_email="uod-author-email",
            uod_filename="uod-filename",
            location="test_location",
            engine_version="1")

        register_response = await self.ctx.engineDispatcher.send_registration_msg_async(register_message)
        assert isinstance(register_response, AM.RegisterEngineReplyMsg)
        self.assertTrue(register_response.success)
        engine_id = register_response.engine_id
        assert engine_id is not None
        self.assertTrue(engine_id != "")
        self.assertEqual(self.ctx.engineDispatcher._engine_id, engine_id)
        self.assertTrue(self.ctx.aggregatorDispatcher.has_connected_engine_id(engine_id))
        return engine_id


class ProtocolIntegrationTest(ProtocolIntegrationTestCase):

    async def test_can_register(self):
        assert self.ctx is not None
        engine_id = await self.register()
        self.assertTrue(engine_id != "")

        engine_data = self.ctx.aggregator.get_registered_engine_data(engine_id)
        assert engine_data is not None

    async def test_can_send_tag_updates(self):
        assert self.ctx is not None

        with database.create_scope():
            await self.ctx.engineRunner._connect_async()
            while self.ctx.engineRunner.state != "Connected":
                await asyncio.sleep(.5)

        assert self.ctx.engineDispatcher._engine_id is not None

        await asyncio.sleep(1)

        engine_data = self.ctx.aggregator.get_registered_engine_data(self.ctx.engineDispatcher._engine_id)
        assert engine_data is not None
        self.assertEqual(len(engine_data.tags_info.map), len(self.ctx.engine.tags))
        self.assertEqual(engine_data.system_state.value, SystemStateEnum.Stopped)  # type: ignore

    async def test_can_detect_network_down_and_buffer_up_messages(self):
        assert self.ctx is not None
        with database.create_scope():
            await self.ctx.engineRunner._connect_async()

        self.ctx.engineDispatcher.network_failing = True
        self.assertTrue(self.ctx.engineRunner._get_buffer_size() == 0)

        # post message while network_failing==True triggers Failed state
        # and the message is buffered
        await self.ctx.engineRunner._post_async(EM.EngineMessage())

        self.assertEqual(self.ctx.engineRunner.state, "Failed")
        self.assertTrue(self.ctx.engineRunner._get_buffer_size() > 0)


class AggregatorErrorTest(ProtocolIntegrationTestCase):
    async def test_can_get_engine_data_for_registered_engine(self):
        # register engine and check engine_data.connected == True
        assert self.ctx is not None
        engine_id = await self.register()

        engine_data = self.ctx.aggregator.get_registered_engine_data(engine_id)
        self.assertIsNotNone(engine_data)

    async def test_rpc_call_fail_with_network_error_engine_state_is_removed(self):
        assert self.ctx is not None

        with database.create_scope():
            await self.ctx.engineRunner._connect_async()
            engine_id = self.ctx.engineDispatcher._engine_id
            assert engine_id is not None

        self.assertTrue(self.ctx.aggregatorDispatcher.has_connected_engine_id(engine_id))

        self.ctx.aggregatorDispatcher.set_engine_fail(engine_id, True)
        try:
            await self.ctx.aggregatorDispatcher.rpc_call(engine_id=engine_id, message=M.MessageBase())
        except Exception:
            pass

        self.assertFalse(self.ctx.aggregatorDispatcher.has_connected_engine_id(engine_id))

        engine_data = self.ctx.aggregator.get_registered_engine_data(engine_id)
        self.assertIsNone(engine_data)


class AggregatorTestDispatcher(AggregatorDispatcher):
    def __init__(self) -> None:
        super().__init__()
        self.engineDispatcher: EngineDispatcher
        self.connected_engines: set[str] = set()
        self.failing_engines: set[str] = set()

    def set_engine_fail(self, engine_id: str, failing: bool):
        self.failing_engines.add(engine_id)

    def get_engine_fail(self, engine_id):
        return engine_id in self.failing_engines

    async def rpc_call(self, engine_id: str, message: M.MessageBase) -> M.MessageBase:
        if self.get_engine_fail(engine_id):
            self.connected_engines.remove(engine_id)
            assert self._disconnect_handler is not None
            await self._disconnect_handler(engine_id)
            raise ProtocolNetworkException()
        if not self.has_connected_engine_id(engine_id):
            raise ProtocolException("Unknown engine: " + engine_id)
        response = await self.engineDispatcher.dispatch_message_async(message)
        assert isinstance(response, M.MessageBase)
        return response

    async def dispatch_message(self, message: EM.EngineMessage) -> M.MessageBase:
        response = await super().dispatch_message(message)
        if isinstance(response, M.ProtocolErrorMessage):
            raise ValueError("Dispatch failed with protocol error: " + str(response.protocol_msg))
        return response

    def has_connected_engine_id(self, engine_id: str) -> bool:
        return engine_id in self.connected_engines

class EngineTestDispatcher(EngineDispatcher):
    def __init__(self) -> None:
        uod_options = {
            'uod_name': 'my_uod_name',
            'uod_author_name': 'my_uod_author_name',
            'uod_author_email': 'my_uod_author_email',
            'uod_filename': 'my_uod_filename',
            'location': 'my_location'
        }
        super().__init__(aggregator_host="", secure=False, uod_options=uod_options)
        self.aggregatorDispatcher: AggregatorTestDispatcher
        self.network_failing = False

    async def connect_async(self):
        if self._engine_id is None:
            logger.info("Registering for engine_id")
            self._engine_id = await self._register_for_engine_id_async()
            if self._engine_id is None:
                logger.error("Failed to register. Aggregator refused registration.")
                raise ValueError("Failed to register. Aggregator refused registration.")

    async def send_registration_msg_async(self, message: EM.RegisterEngineMsg) -> M.MessageBase:
        assert self.aggregatorDispatcher is not None
        assert self.aggregatorDispatcher._register_handler is not None
        result = await self.aggregatorDispatcher._register_handler(message)
        assert isinstance(result, AM.RegisterEngineReplyMsg)
        assert result.engine_id is not None
        self.aggregatorDispatcher.connected_engines.add(result.engine_id)
        self._engine_id = result.engine_id
        return result

    async def send_async(self, message: EM.EngineMessage) -> M.MessageBase:
        # fill in message.engine_id
        if (self._engine_id is None):
            raise Exception("_engine_id should have been set")
        message.engine_id = self._engine_id
        if self.network_failing:
            raise ProtocolNetworkException
        # post message directly to aggregatorDispatcher
        response = await self.aggregatorDispatcher.dispatch_message(message)
        if isinstance(response, AM.RegisterEngineReplyMsg) and response.success:
            self._engine_id = response.engine_id
        return response


class MessageTest(unittest.TestCase):
    def test_print_sequence_range(self):
        self.assertEqual("", EM.print_sequence_range([]))

        m1 = EM.EngineMessage(sequence_number=1)
        self.assertEqual("1", EM.print_sequence_range([m1]))

        m2 = EM.EngineMessage(sequence_number=2)
        m3 = EM.EngineMessage(sequence_number=3)
        self.assertEqual("1-3", EM.print_sequence_range([m1, m2, m3]))
        self.assertEqual("2-3", EM.print_sequence_range([m2, m3]))
        self.assertEqual("1,3", EM.print_sequence_range([m1, m3]))

        m4 = EM.EngineMessage(sequence_number=4)
        m5 = EM.EngineMessage(sequence_number=5)
        self.assertEqual("1-5", EM.print_sequence_range([m1, m2, m3, m4, m5]))
        self.assertEqual("1-2,4-5", EM.print_sequence_range([m1, m2, m4, m5]))
        self.assertEqual("1-3,5", EM.print_sequence_range([m1, m2, m3, m5]))
        self.assertEqual("2,4-5", EM.print_sequence_range([m2, m4, m5]))


if __name__ == "__main__":
    unittest.main()



import logging
import ctypes
import os
import sys
import json
import functools
import re
from typing import Any, Sequence, Optional, Dict

from openpectus.engine.hardware import (
    HardwareLayerBase,
    Register,
    RegisterDirection,
    HardwareLayerException,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class SuppressPrint:
    """ Supress print
    Source: https://stackoverflow.com/questions/8391411/how-to-block-calls-to-print """
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


# The LabJack library executes code on import and this cannot be hindered.
# The code attempts to load the LabJack DLL but fails if it is not installed.
# This is not a problem, because the DLL is shipped with this sofware.
# However, the error message is still printed and this may confuse the user.
with SuppressPrint():
    import labjack.ljm.ljm as ljm
    import labjack.ljm.constants


DEVICE_TYPES = {
    labjack.ljm.constants.dtT4: "T4",
    labjack.ljm.constants.dtT7: "T7",
    labjack.ljm.constants.dtT8: "T8",
}

CONNECTION_TYPES = {
    labjack.ljm.constants.ctANY: "Any",
    labjack.ljm.constants.ctUSB: "USB",
    labjack.ljm.constants.ctTCP: "TCP",
    labjack.ljm.constants.ctETHERNET: "Ethernet",
    labjack.ljm.constants.ctWIFI: "Wifi",
    labjack.ljm.constants.ctNETWORK_UDP: "UDP",
    labjack.ljm.constants.ctETHERNET_UDP: "Ethernet UDP",
    labjack.ljm.constants.ctWIFI_UDP: "Wifi UDP",
    labjack.ljm.constants.ctNETWORK_ANY: "Network",
    labjack.ljm.constants.ctETHERNET_ANY: "Ethernet",
    labjack.ljm.constants.ctWIFI_ANY: "Wifi",
    labjack.ljm.constants.ctANY_UDP: "UDP",
}

# The Labjack DLL/so are not distributed with the python package.
# They are included here along with the obligatory LICENSE.
if sys.platform.startswith("win32"):
    ljm._staticLib = ctypes.CDLL(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "labjack",
            "LabJackM.dll"))
elif sys.platform.startswith("linux"):
    ljm._staticLib = ctypes.CDLL(
        os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "labjack",
            "libLabJackM.so"))

constants_file = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "labjack",
    "ljm_constants.json")
ljm.loadConstantsFromFile(constants_file)


class Labjack_Hardware(HardwareLayerBase):
    """ Represents LabJack hardware layer. """
    def __init__(self, serial_number: Optional[str] = None) -> None:
        super().__init__()
        self.serial_number: str = serial_number if serial_number else "ANY"
        self._handle = None

    @functools.cached_property
    def port_directions(self):
        # Load ljm_constants.json
        with open(constants_file, "r") as f:
            ljm_constants = json.load(f)
        # Prepare regex and result dict
        regex = re.compile(r"#\((?P<start>\d+):(?P<end>\d+)\)")
        ljm_register_directions: Dict[str, RegisterDirection] = dict()
        direction_to_register_direction = {"R": RegisterDirection.Read,
                                           "W": RegisterDirection.Write,
                                           "RW": RegisterDirection.Both}
        # Note the port direction for each register
        for register in ljm_constants["registers"]:
            name = register["name"]
            # Some register names cover ranges of registers.
            # Example: "AIN#(0:254)"
            match = re.search(regex, name)
            if match:
                match_dict = match.groupdict()
                start, end = int(match_dict["start"]), int(match_dict["end"])
                base_name = name.replace(match.group(0), "{i}")
                names = [base_name.format(i=i) for i in range(start, end+1)]
            else:
                names = [name]
            for name in names:
                ljm_register_directions[name] = direction_to_register_direction[register["readwrite"]]
        return ljm_register_directions

    def validate_offline(self):
        for r in self.registers.values():
            if "port" not in r.options:
                raise HardwareLayerException((f"Labjack hardware layer requires specification of "
                                              f"port for register {r}."))
            # Check that the register port exists
            port = r.options["port"]
            try:
                address, data_type = ljm.nameToAddress(port)
            except ljm.LJMError as error:
                raise HardwareLayerException(f"Register {r} port '{port}' invalid. Labjack error: {error}.")
            # Check the direction
            labjack_port_direction = self.port_directions[port]
            if (
               RegisterDirection.Read in r.direction and
               RegisterDirection.Read not in labjack_port_direction
               ) or (
               RegisterDirection.Write in r.direction and
               RegisterDirection.Write not in labjack_port_direction
               ):
                raise HardwareLayerException((f"Disparity between register direction '{r.direction}' "
                                              f"for register {r} and Labjack port direction "
                                              f"'{labjack_port_direction}."))

    def read(self, r: Register) -> Any:
        if RegisterDirection.Read not in r.direction:
            raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        try:
            return ljm.eReadName(self._handle, r.options["port"])
        except ljm.LJMError as error:
            logger.error(f"Unable to read {r}", exc_info=True)
            raise HardwareLayerException(f"Unable to read {r}. Labjack error: {error}.")

    def read_batch(self, registers: Sequence[Register]) -> list[Any]:
        """ Read batch of register values with a single OPC-UA call. """
        for r in registers:
            if RegisterDirection.Read not in r.direction:
                raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        try:
            return ljm.eReadNames(self._handle, len(registers), [r.options["port"] for r in registers])
        except ljm.LJMError as error:
            logger.error(f"Unable to read registers {registers}", exc_info=True)
            raise HardwareLayerException(f"Unable to read registers {registers}. Labjack error: {error}.")

    def write(self, value: Any, r: Register):
        if RegisterDirection.Write not in r.direction:
            raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        try:
            return ljm.eWriteName(self._handle, r.options["port"], value)
        except ljm.LJMError as error:
            logger.error(f"Unable to write '{value}' to {r}", exc_info=True)
            raise HardwareLayerException(f"Unable to write '{value}' to {r}. Labjack error: {error}.")

    def write_batch(self, values: Sequence[Any], registers: Sequence[Register]):
        """ Write batch of register values with a single OPC-UA call. """
        for r in registers:
            if RegisterDirection.Write not in r.direction:
                raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        try:
            ljm.eWriteNames(self._handle, len(registers), [r.options["port"] for r in registers], values)
        except ljm.LJMError as error:
            logger.error("Unable to write batch", exc_info=True)
            raise HardwareLayerException(f"Unable to write registers {registers}. Labjack error: {error}.")

    def setup(self):
        """ Method to call after successful connect. """
        pass

    def _reconnect(self, handle):
        logger.info(f"Reconnected to Labjack with serial number {self.serial_number}")

    def connect(self):
        """ Connect to Labjack. """
        logger.info(f"Attempting to connect to Labjack with serial number {self.serial_number}")
        if self._handle:
            self.disconnect()
        try:
            self._handle = ljm.openS("ANY",  # Device (T4, T7, T8)
                                     "ANY",  # Connection (USB, ETHERNET)
                                     self.serial_number)  # Identifier (Serial number)
        except ljm.LJMError:
            logger.info(f"Unable to connect to Labjack with serial number {self.serial_number}")
            # Look for devices
            n, device_types, connection_types, serial_numbers, ip_addresses = ljm.listAllS("ANY", "ANY")
            if n > 0:
                logger.info(f"Found {n} Labjack device(s):")
                logger.info(" Type    Connection      Serial Number      IP")
                for (device_type,
                     connection_type,
                     serial_number,
                     ip_addresse) in zip(device_types,
                                         connection_types,
                                         serial_numbers,
                                         ip_addresses):
                    logger.info((f"  {DEVICE_TYPES[device_type]:<10} "
                                 f"{CONNECTION_TYPES[connection_type]:<12}  "
                                 f"{serial_number:<10}       "
                                 f"{ip_addresse:<15}"))
            raise HardwareLayerException(f"Unable to connect to Labjack with serial number {self.serial_number}")
        logger.info(f"Connected to {self.serial_number}")
        self.setup()
        ljm.registerDeviceReconnectCallback(self._handle, self._reconnect)
        super().connect()

    def disconnect(self):
        """ Disconnect hardware. """
        logger.info(f"Disconnecting from Labjack with serial number {self.serial_number}")
        if self._handle:
            ljm.close(self._handle)
            self._handle = None
        super().disconnect()

    def __str__(self):
        return f"Labjack_Hardware(serial_number={self.serial_number})"



import openpectus.protocol.messages as Msg
import openpectus.protocol.models as Mdl

# Note: These are the messages sent by AggregatorDispatcher and handled by EngineMessageHandlers.


SuccessMessage = Msg.SuccessMessage
ErrorMessage = Msg.ErrorMessage


class AggregatorMessage(Msg.MessageBase):
    """ Marker class that indicates a protocol message sent by Aggregator to Engine.

    On the Engine side, they are handled by EngineMessageHandlers.
    """
    pass


class RegisterEngineReplyMsg(AggregatorMessage):
    """ Reply to a RegisterEngineMsg from engine. """
    success: bool
    engine_id: str | None = None
    secret_match: bool


class InvokeCommandMsg(AggregatorMessage):
    """ Request from user to invoke a command. """
    name: str = ""
    arguments: str | None = None


class InjectCodeMsg(AggregatorMessage):
    """ Request from user to inject pcode. """
    pcode: str


class MethodMsg(AggregatorMessage):
    """ Request from user to set a new method to be executed by engine. """
    method: Mdl.Method


class CancelMsg(AggregatorMessage):
    """ Request from user to cancel an instruction. """
    exec_id: str


class ForceMsg(AggregatorMessage):
    """ Request from user to force an instruction to start. """
    exec_id: str



from __future__ import annotations
import logging
import time
from typing import Any
import unittest

from openpectus.engine.hardware import (
    HardwareLayerBase,
    HardwareLayerException,
    Register,
    RegisterDirection,
)
from openpectus.engine.hardware_recovery import (
    ErrorRecoveryConfig,
    ErrorRecoveryDecorator,
    ErrorRecoveryState,
)
from openpectus.engine.models import ConnectionStatusEnum
from openpectus.lang.exec.tags import SystemTagName, TagCollection
from openpectus.test.engine.test_engine import TestHW, create_test_uod


FORMAT = "%(asctime)-15s %(message)s"
logging.basicConfig(format='%(asctime)-15s :: %(name)s :: %(levelname)-8s :: %(message)s', datefmt='%Y-%m-%d %H:%M:%S',
                    force=True)
logger = logging.getLogger("openpectus.engine.hardware_error")
logger.setLevel(logging.DEBUG)


class TestHardwareLayer(unittest.TestCase):
    def test_can_read_register(self):
        uod = create_test_uod()
        hwl: TestHW = uod.hwl  # type: ignore
        self.assertIsInstance(hwl, TestHW)
        self.assertIsInstance(hwl, HardwareLayerBase)

        rFT01 = hwl.registers["FT01"]

        self.assertEqual(None, hwl.read(rFT01))
        hwl.register_values["FT01"] = 78

        self.assertEqual(78, hwl.read(rFT01))

    def test_can_write_register(self):
        uod = create_test_uod()
        hwl: TestHW = uod.hwl  # type: ignore

        rFT01 = hwl.registers["FT01"]

        hwl.write("foo", rFT01)

        self.assertEqual("foo", hwl.register_values["FT01"])


class TestHardwareErrorRecovery(unittest.TestCase):

    def create_hardwares(self) -> tuple[ErrorRecoveryDecorator, ErrorTestHardware]:
        hwl = ErrorTestHardware()
        error_config = ErrorRecoveryConfig()
        connection_status_tag = TagCollection.create_system_tags()[SystemTagName.CONNECTION_STATUS]
        return ErrorRecoveryDecorator(hwl, error_config, connection_status_tag), hwl

    def test_initial_state_Disconnected(self):
        decorator, hwl = self.create_hardwares()
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Disconnected)
        self.assertEqual(decorator.connection_status_tag.get_value(), ConnectionStatusEnum.Disconnected)

    def test_state_Disconnected_connect_ok_Connected(self):
        decorator, hwl = self.create_hardwares()
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Disconnected)

        decorator.connect()

        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.OK)
        self.assertEqual(decorator.connection_status_tag.get_value(), ConnectionStatusEnum.Connected)

    def test_initialize_with_connected_hwl(self):
        hwl = ErrorTestHardware()
        hwl._is_connected = True
        error_config = ErrorRecoveryConfig()
        connection_status_tag = TagCollection.create_system_tags()[SystemTagName.CONNECTION_STATUS]
        decorator = ErrorRecoveryDecorator(hwl, error_config, connection_status_tag)

        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.OK)
        self.assertEqual(decorator.is_connected, True)

    def test_state_Disconnected_connect_error_Disconnected(self):
        decorator, hwl = self.create_hardwares()
        hwl.connect_fail = True

        with self.assertRaises(HardwareLayerException):
            decorator.connect()

        self.assertEqual(decorator.connection_status_tag.get_value(), ConnectionStatusEnum.Disconnected)

    def test_state_Disconnected_read_raises(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Disconnected
        reg_A = hwl.reg_A

        with self.assertRaises(HardwareLayerException):
            decorator.read(reg_A)
        with self.assertRaises(HardwareLayerException):
            decorator.read_batch([reg_A])

    def test_state_Error_read_raises(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Error
        reg_A = hwl.reg_A

        with self.assertRaises(HardwareLayerException):
            decorator.read(reg_A)
        with self.assertRaises(HardwareLayerException):
            decorator.read_batch([reg_A])

    def test_read_error_returns_last_known_good_value_masking_error(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()

        reg_A = hwl.reg_A
        hwl.reg_A_value = 3
        val_A = decorator.read(reg_A)
        self.assertEqual(val_A, 3)

        hwl.read_fail = True
        val_A_fail = decorator.read(reg_A)
        self.assertEqual(val_A_fail, 3)

    def test_read_error_in_state_OK_transitions_to_state_Issue(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.OK)

        hwl.read_fail = True
        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Issue)

    def test_read_error_in_state_Issue_transitions_to_state_Reconnect(self):
        decorator, hwl = self.create_hardwares()
        decorator.config.reconnect_timeout_seconds = 1
        decorator.config.error_timeout_seconds = 2
        decorator.connect()
        _ = decorator.read(hwl.reg_A)  # make last_known_good value available
        hwl.read_fail = True

        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Issue)

        time.sleep(1)

        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Reconnect)
        self.assertEqual(decorator.connection_status_tag.get_value(), ConnectionStatusEnum.Connected)

    def test_read_in_state_Reconnect_transitions_to_state_Error_after_timeout(self):
        decorator, hwl = self.create_hardwares()
        decorator.config.reconnect_timeout_seconds = 1
        decorator.config.error_timeout_seconds = 1
        decorator.connect()
        _ = decorator.read(hwl.reg_A)  # make last_known_good value available
        hwl.read_fail = True

        _ = decorator.read(hwl.reg_A)
        time.sleep(1)
        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Reconnect)

        time.sleep(1)

        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Error)

    def test_read_success_in_state_Issue_transitions_to_state_OK(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        _ = decorator.read(hwl.reg_A)  # make last_known_good value available
        hwl.read_fail = True

        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Issue)

        hwl.read_fail = False
        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.OK)

    def test_read_success_in_state_Reconnect_keeps_state_Reconnect(self):
        decorator, hwl = self.create_hardwares()
        decorator.config.reconnect_timeout_seconds = 1
        decorator.config.error_timeout_seconds = 1
        decorator.connect()
        _ = decorator.read(hwl.reg_A)  # make last_known_good value available
        hwl.read_fail = True

        _ = decorator.read(hwl.reg_A)
        time.sleep(1)
        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Reconnect)

        hwl.read_fail = False
        _ = decorator.read(hwl.reg_A)

        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Reconnect)

    # Writes

    def test_state_Disconnected_write_raises(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Disconnected
        reg_B = hwl.reg_B

        with self.assertRaises(HardwareLayerException):
            decorator.write(3, reg_B)
        with self.assertRaises(HardwareLayerException):
            decorator.write_batch([4], [reg_B])

    def test_state_Error_write_raises(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Error
        reg_B = hwl.reg_B

        with self.assertRaises(HardwareLayerException):
            decorator.write(3, reg_B)
        with self.assertRaises(HardwareLayerException):
            decorator.write_batch([4], [reg_B])


    def test_failing_write_masks_error_and_adds_to_pending(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.OK
        reg_B = hwl.reg_B
        hwl.reg_B_value = 2

        hwl.write_fail = True
        decorator.write(3, reg_B)  # error masked
        self.assertEqual(hwl.reg_B_value, 2)  # value not written
        self.assertEqual(decorator.pending_writes[reg_B], 3)  # value added to pending


    def test_write_ok_in_state_Issue_writes_pending_values(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Issue
        hwl.reg_B_value = 2
        decorator.pending_writes[hwl.reg_B] = 3

        # write succcesfully to another register
        decorator.write(7, hwl.reg_C)

        # applies the pending write to B
        self.assertEqual(hwl.reg_B_value, 3)

    # Reconnect

    def test_reconnect_is_correctly_backed_off(self):
        decorator, hwl = self.create_hardwares()
        decorator.connect()
        decorator.state = ErrorRecoveryState.Reconnect

        # collect the ticks for which reconnect is called
        hwl.connect_fail = True  # make all reconnects fail
        reconnected_ticks = []
        decorator.reconnecting_callback = lambda : reconnected_ticks.append(decorator.reconnect_tick)

        largest = decorator.reconnect_backoff_ticks[-1]
        for _ in range(0, 2 * (largest + 1)):
            decorator.tick()

        expected_ticks = decorator.reconnect_backoff_ticks + [2 * largest]

        # because all reconnects fail, we collected all the backoff ticks
        self.assertEqual(reconnected_ticks, expected_ticks)

    def test_in_state_Reconnect_reconnect_is_automatically_performed_on_tick(self):
        decorator, hwl = self.create_hardwares()
        decorator.config.reconnect_timeout_seconds = 1
        decorator.config.error_timeout_seconds = 5
        decorator.connect()
        _ = decorator.read(hwl.reg_A)  # make last_known_good value available
        hwl.read_fail = True

        is_reconnecting = False

        def on_reconnecting():
            nonlocal is_reconnecting
            is_reconnecting = True
        decorator.on_reconnecting = on_reconnecting

        _ = decorator.read(hwl.reg_A)
        time.sleep(1)
        _ = decorator.read(hwl.reg_A)
        self.assertEqual(decorator.get_recovery_state(), ErrorRecoveryState.Reconnect)

        self.assertEqual(is_reconnecting, False)

        for _ in range(1 + decorator.reconnect_backoff_ticks[0]):
            decorator.tick()

        self.assertEqual(is_reconnecting, True)

    def test_registers_are_delegated(self):
        decorator, hwl = self.create_hardwares()
        self.assertEqual(hwl.reg_A, decorator.registers["A"])
        self.assertEqual(len(hwl.registers), len(decorator.registers))

    def test_methods_are_delegated(self):
        decorator, hwl = self.create_hardwares()
        # Since test hardware has this method, it should also be callable on the decorator. This is done using
        # method forwarding, implemented in ErrorRecoveryDecorator._setup_base_method_forwards()
        decorator.test_cmd()  # type: ignore

class ErrorTestHardware(HardwareLayerBase):
    __test__ = False
    """ A test hardware class that can fail its operations when so directed by the test"""
    def __init__(self) -> None:
        super().__init__()

        self.reg_A = self.registers["A"] = Register("A", RegisterDirection.Read)
        self.reg_B = self.registers["B"] = Register("B", RegisterDirection.Write)
        self.reg_C = self.registers["C"] = Register("C", RegisterDirection.Write)
        self.read_fail = False
        self.write_fail = False
        self.connect_fail = False
        self.disconnect_fail = False

        self.reg_A_value = 0
        self.reg_B_value = 0
        self.reg_C_value = 0

    def test_cmd(self):
        print("Test Cmd executed")

    def read(self, r: Register) -> Any:
        if self.read_fail:
            raise HardwareLayerException("Read failed as requested")
        if r.name == "A":
            return self.reg_A_value
        else:
            raise ValueError("Cannot read from register " + r.name)

    def write(self, value: Any, r: Register) -> None:
        if self.write_fail:
            raise HardwareLayerException("Write failed as requested")
        if r.name == "B":
            self.reg_B_value = value
        elif r.name == "C":
            self.reg_C_value = value
        else:
            raise ValueError("Cannot write to register " + r.name)

    def connect(self):
        if self.connect_fail:
            raise HardwareLayerException("Connect failed as requested")
        self._is_connected = True

    def disconnect(self):
        # The order here is tricky. If it fails, we cannot reasonably know
        # whether self._is_connected is set and to what value. But when
        # we are using the decorator we don't use this value because the
        # decorator knows better.
        self._is_connected = False
        if self.disconnect_fail:
            raise HardwareLayerException("Disconnect failed as requested")


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import itertools
import logging
import uuid
from queue import Empty, Queue
from typing import Iterable, List, Set
from uuid import UUID
from openpectus.engine.internal_commands import InternalCommandsRegistry
from openpectus.engine.hardware import HardwareLayerException, RegisterDirection
from openpectus.engine.method_model import MethodModel
from openpectus.engine.models import MethodStatusEnum, SystemStateEnum, EngineCommandEnum, SystemTagName
from openpectus.lang.exec.base_unit import BaseUnitProvider
from openpectus.lang.exec.clock import Clock, WallClock
from openpectus.lang.exec.commands import CommandRequest
from openpectus.lang.exec.errors import (
    EngineError, InterpretationError, InterpretationInternalError
)
from openpectus.lang.exec.pinterpreter import PInterpreter, InterpreterContext
from openpectus.lang.exec.runlog import RuntimeInfo, RunLog, RuntimeRecord
from openpectus.lang.exec.events import EventEmitter
from openpectus.lang.exec.tags import (
    Tag,
    TagCollection,
    TagDirection,
    TagValue,
    TagValueCollection,
    ChangeListener,
    Unset,
)
from openpectus.lang.exec.tags_impl import MarkTag
from openpectus.lang.exec.timer import EngineTimer, OneThreadTimer
from openpectus.lang.exec.uod import UnitOperationDefinitionBase, UodCommand
from openpectus.lang.grammar.pgrammar import PGrammar
from openpectus.lang.model.pprogram import PProgram
import openpectus.protocol.models as Mdl
from openpectus.engine.archiver import ArchiverTag

logger = logging.getLogger(__name__)
frontend_logger = logging.getLogger(__name__ + ".frontend")

class EngineTiming():
    """ Represents timing information used by engine and any related components. """
    def __init__(self, clock: Clock, timer: EngineTimer, interval: float, speed: float) -> None:
        self._clock = clock
        self._timer = timer
        self._interval = interval
        self._speed = speed

    @staticmethod
    def default() -> EngineTiming:
        return EngineTiming(WallClock(), OneThreadTimer(0.1, None), 0.1, 1.0)

    @property
    def clock(self) -> Clock:
        return self._clock

    @property
    def timer(self) -> EngineTimer:
        return self._timer

    @property
    def interval(self) -> float:
        return self._interval

    @property
    def speed(self) -> float:
        return self._speed

class Engine(InterpreterContext):
    """ Main engine class. Handles
    - io loop, reads and writes hardware process image (sync)
    - invokes interpreter to interpret next instruction (sync, generator based)
    - signals state changes via tag_updates queue (to EngineReporter)
    - accepts commands from cmd_queue (from interpreter and from aggregator)
    """

    def __init__(
            self,
            uod: UnitOperationDefinitionBase,
            timing: EngineTiming = EngineTiming.default(),
            enable_archiver=False) -> None:
        self.uod = uod
        self._running: bool = False
        """ Indicates whether the scan cycle loop is running, set to False to shut down"""

        self.registry = InternalCommandsRegistry(self).__enter__()
        """ Manages internal engine commands """

        self._clock: Clock = timing.clock
        """ The time source """

        self._tick_timer: EngineTimer = timing.timer
        """ Timer that invokes tick() """

        self._tick_time: float = 0.0
        """ The time of the last tick """
        self._tick_number: int = -1
        """ Tick count since last START command. It is incremented at the start of each tick.
        First tick is effectively number 0. """

        self._system_tags = TagCollection.create_system_tags()
        self._system_tags.add(MarkTag())
        # Add archiver which is implemented as a tag. The lambda getting the runlog works because the
        # tag_lifetime.on_stop event is emitted just before resetting the interpreter and runlog (and
        # not after).
        if enable_archiver:
            archiver = ArchiverTag(
                lambda : self.runtimeinfo.get_runlog(),
                lambda : self.tags,
                self.uod.data_log_interval_seconds)
            self._system_tags.add(archiver)

        self.uod.system_tags = self._system_tags

        self.cmd_queue: Queue[CommandRequest] = Queue()
        """ Commands to execute, coming from interpreter and from aggregator """
        self.cmd_executing: List[CommandRequest] = []
        """ Uod commands currently being excuted """
        self.tag_updates: Queue[Tag] = Queue()
        """ Tags updated in last tick """

        self._uod_listener = ChangeListener()
        self._system_listener = ChangeListener()

        self._runstate_started: bool = False
        """ Indicates the current Start/Stop state"""
        self._runstate_started_time: float = 0
        """ Indicates the time of the last Start state"""
        self._runstate_paused: bool = False
        """ Indicates whether the engine is paused"""
        self._runstate_holding: bool = False
        """ Indicates whether the engine is on hold"""
        self._runstate_stopping: bool = False
        """ Indicates whether the engine is on stopping"""
        self._runstate_waiting: bool = False
        """ Indicates whether the engine is waiting in a Wait command"""

        self._prev_state: TagValueCollection | None = None
        """ The state prior to applying safe state """

        self._last_error : Exception | None = None
        """ Cause of error_state"""

        self.block_times: dict[str, float] = {}
        """ holds time spent in each block"""

        self._interpreter: PInterpreter = PInterpreter(PProgram(), self)
        """ The interpreter executing the current program. """

        def on_method_init():
            self._interpreter.stop()
            self._interpreter = PInterpreter(self._method.get_program(), self)
            logger.info("Method and interpreter re-initialized")

        def on_method_error(ex: Exception):
            logger.error("An error occured while setting new method: " + str(ex))

        self._method: MethodModel = MethodModel(on_method_init, on_method_error)
        """ The model handling changes to program/method code """

        self._cancel_command_exec_ids: set[UUID] = set()

        # initialize state
        self.uod.tags.add_listener(self._uod_listener)
        self._system_tags.add_listener(self._system_listener)
        self._tags = self._system_tags.merge_with(self.uod.tags)
        self._emitter = EventEmitter(self._tags)
        self._tick_timer.set_tick_fn(self.tick)


    def _iter_all_tags(self) -> Iterable[Tag]:
        return itertools.chain(self._system_tags, self.uod.tags)

    def tags_as_readonly(self) -> TagValueCollection:
        return TagValueCollection(t.as_readonly() for t in self._iter_all_tags())

    @property
    def tags(self) -> TagCollection:
        return self._tags

    @property
    def emitter(self) -> EventEmitter:
        return self._emitter

    @property
    def base_unit_provider(self) -> BaseUnitProvider:
        return self.uod.base_unit_provider

    @property
    def interpreter(self) -> PInterpreter:
        assert self._interpreter is not None
        return self._interpreter

    @property
    def runtimeinfo(self) -> RuntimeInfo:
        return self.interpreter.runtimeinfo

    @property
    def method(self) -> MethodModel:
        return self._method

    def cleanup(self):
        self.emitter.emit_on_engine_shutdown()
        self.registry.__exit__(None, None, None)

    def get_runlog(self) -> RunLog:
        return self.runtimeinfo.get_runlog()

    def run(self, skip_timer_start=False):
        self._run(skip_timer_start)

    def is_running(self) -> bool:
        return self._running

    def _run(self, skip_timer_start=True):
        """ Starts the scan cycle """

        assert self.uod is not None
        assert self.uod.hwl is not None
        assert self.uod.hwl.is_connected, "Hardware is not connected. Engine cannot start"

        self._running = True
        if not skip_timer_start:
            self._tick_timer.start()

        self.emitter.emit_on_engine_configured()

        # On engine start, write safe output values to hardware to bring hw to a known state
        self._apply_safe_state()
        self.write_process_image()

    def stop(self):
        logger.info("Engine shutting down")
        try:
            self.uod.hwl.disconnect()
        except HardwareLayerException:
            logger.error("Disconnect hardware error", exc_info=True)

        self._running = False
        self._tick_timer.stop()
        self.cleanup()

    def tick(self, tick_time: float, increment_time: float):
        """ Performs a scan cycle tick. """
        # logger.debug(f"Tick {self._tick_number + 1}")

        if not self._running:
            self._tick_timer.stop()

        self._tick_time = tick_time
        self._tick_number += 1

        # Perform certain actions in first tick
        if self._tick_number == 0:
            # System tags are initialized before first tick, without a tick time, and some are never updated, so
            # provide first tick time as a "default".
            for tag in self._system_tags.tags.values():
                tag.tick_time = tick_time

        self.uod.hwl.tick()

        # read, error_state on HardwareLayerException
        self.read_process_image()

        # excecute interpreter tick
        if self._runstate_started and\
                not self._runstate_paused and\
                not self._runstate_holding and\
                not self._runstate_waiting and\
                not self._runstate_stopping:
            try:
                # run one tick of interpretation, i.e. one instruction
                self._interpreter.tick(tick_time, self._tick_number)
            except InterpretationInternalError as ex:
                logger.fatal("A serious internal interpreter error occured. The method should be stopped. If it is resumed, \
                             additional errors may occur.", exc_info=True)
                self.set_error_state(ex)
            except EngineError as eex:
                logger.error(eex.message)
                if eex.user_message is not None:
                    frontend_logger.error(eex.user_message)
                self.set_error_state(eex)
            except InterpretationError as ie:
                logger.error("Interpretation error", exc_info=True)
                if ie.user_message is not None:
                    frontend_logger.error(ie.user_message)
                self.set_error_state(ie)
            except Exception as ex:
                logger.error("Unhandled interpretation error", exc_info=True)
                frontend_logger.error("Method error")
                self.set_error_state(ex)

        # update calculated tags
        if self._runstate_started:
            self.update_calculated_tags(tick_time, increment_time)

        # execute queued commands, go to error_state on error
        self.execute_commands()

        # notify of tag changes
        self.notify_tag_updates()

        # write, error_state on HardwareLayerException
        self.write_process_image()

    def read_process_image(self):
        """ Read data from relevant hw registers into tags"""
        read_registers = [r for r in self.uod.hwl.registers.values() if RegisterDirection.Read in r.direction]
        try:
            # possibly add guard for ConnectionState=Disconnected to avoid read errors
            # flooding the log
            register_values = self.uod.hwl.read_batch(read_registers)
        except HardwareLayerException as ex:
            if not self.has_error_state():
                logger.error("Hardware read_batch error", exc_info=True)
                self.set_error_state(ex)
            return

        for i, r in enumerate(read_registers):
            tag = self.uod.tags.get(r.name)
            tag_value = register_values[i]
            if "to_tag" in r.options:
                tag_value = r.options["to_tag"](tag_value)
            tag.set_value(tag_value, self._tick_time)

    def update_calculated_tags(self, tick_time: float, increment_time: float):
        sys_state = self._system_tags[SystemTagName.SYSTEM_STATE]
        logger.debug(f"{increment_time=}")

        # Clock         - seconds since epoch
        clock = self._system_tags.get(SystemTagName.CLOCK)
        clock.set_value(tick_time, tick_time)

        # Process Time  - 0 at start, increments when System State is Run
        process_time = self._system_tags[SystemTagName.PROCESS_TIME]
        process_time_value = process_time.as_number()
        if sys_state.get_value() == SystemStateEnum.Running:
            process_time.set_value(process_time_value + increment_time, tick_time)

        # Run Time      - 0 at start, increments when System State is not Stopped
        run_time = self._system_tags[SystemTagName.RUN_TIME]
        run_time_value = run_time.as_number()
        if sys_state.get_value() not in [SystemStateEnum.Stopped, SystemStateEnum.Restarting]:
            run_time.set_value(run_time_value + increment_time, tick_time)

        # Block name + signal block changes to tag_context
        block_name = self._system_tags[SystemTagName.BLOCK].get_value() or ""
        assert isinstance(block_name, str)

        # Block Time    - 0 at Block start, global but value refers to active block
        if block_name not in self.block_times.keys():
            self.block_times[block_name] = 0.0
        # TODO should probably increment all parent block timers as well.
        # In that case factor this out into BlockTag class that hold timer tags for each block
        self.block_times[block_name] += increment_time
        self._system_tags[SystemTagName.BLOCK_TIME].set_value(self.block_times[block_name], self._tick_time)

        # Execute the tick lifetime hook on tags
        self.emitter.emit_on_tick(tick_time, increment_time)

    def execute_commands(self):
        done = False
        # add command request from incoming queue

        while self.cmd_queue.qsize() > 0 and not done:
            try:
                engine_command = self.cmd_queue.get()
                # Note: New commands are inserted at the beginning of the list.
                # This allows simpler cancellation of identical/overlapping commands
                self.cmd_executing.insert(0, engine_command)
            except Empty:
                done = True

        # execute a tick of each running command
        cmds_done: Set[CommandRequest] = set()
        latest_cmd = "(none)"
        try:
            for c in self.cmd_executing:
                latest_cmd = c.name
                if c not in cmds_done:
                    # Note: Executing one command may cause other commands to be cancelled (by identical or overlapping
                    # commands) Rather than modify self.cmd_executing (while iterating over it), cancelled/completed
                    # commands are added to the cmds_done set.
                    self._execute_command(c, cmds_done)
        except ValueError as ve:
            logger.error(f"Error executing command: '{latest_cmd}'. Command failed with error: {ve}", exc_info=True)
            frontend_logger.error(f"Command '{latest_cmd}' failed: {ve}")
            self.set_error_state(ve)
        except Exception as ex:
            logger.error(f"Error executing command: '{latest_cmd}'", exc_info=True)
            frontend_logger.error(f"Error executing command: '{latest_cmd}'")
            self.set_error_state(ex)
        finally:
            for c_done in cmds_done:
                self.cmd_executing.remove(c_done)

    def _execute_command(self, cmd_request: CommandRequest, cmds_done: Set[CommandRequest]):
        # execute an internal engine command or a uod command

        logger.debug("Executing command: " + cmd_request.name)
        if cmd_request.name is None or len(cmd_request.name.strip()) == 0:
            logger.error("Command name empty")
            frontend_logger.error("Cannot execute command with empty name")
            cmds_done.add(cmd_request)
            return

        if EngineCommandEnum.has_value(cmd_request.name):
            self._execute_internal_command(cmd_request, cmds_done)
        else:
            self._execute_uod_command(cmd_request, cmds_done)

    def _execute_internal_command(self, cmd_request: CommandRequest, cmds_done: Set[CommandRequest]):

        if not self._runstate_started and cmd_request.name not in [EngineCommandEnum.START, EngineCommandEnum.RESTART]:
            logger.warning(f"Command {cmd_request.name} is invalid when Engine is not running")
            cmds_done.add(cmd_request)
            return

        # get the runtime record to use for tracking if possible
        record = RuntimeRecord.null_record()
        if cmd_request.exec_id is not None:  # happens for all commands not originating from interpreter
            # during restart, record is None - should ont occur otherwise
            record = self.interpreter.runtimeinfo.get_exec_record(cmd_request.exec_id)

        # an existing, long running engine_command is running. other commands must wait
        # Note: we need a priority mechanism - even Stop is waiting here
        command = self.registry.get_running_internal_command()
        if command is not None:
            if cmd_request.name == command.name:
                if not command.is_finalized():
                    command.tick()
                if command.has_failed():
                    cmds_done.add(cmd_request)
                    if record is not None:
                        record.add_state_failed(self._tick_time, self._tick_number, self.tags_as_readonly())
                    else:
                        logger.error(f"Failed to record failed state for command {cmd_request}")
                elif command.is_finalized():
                    cmds_done.add(cmd_request)
                    if record is not None:
                        record.add_state_completed(self._tick_time, self._tick_number, self.tags_as_readonly())
                    else:
                        logger.error(f"Failed to record completed state for command {cmd_request}")
            return

        # no engine command is running - start one
        try:
            command = self.registry.create_internal_command(cmd_request.name)
            args = cmd_request.get_args()
            if args is not None:
                try:
                    command.init_args(args)
                    logger.debug(f"Initialized command {cmd_request.name} with arguments {args}")
                except Exception:
                    raise EngineError(
                        f"Failed to initialize arguments '{args}' for command '{cmd_request.name}'",
                        "same"
                    )

        except ValueError:
            raise EngineError(
                f"Unknown internal engine command '{cmd_request.name}'",
                f"Unknown command '{cmd_request.name}'")

        if record is not None:
            record.add_state_started(self._tick_time, self._tick_number, self.tags_as_readonly())
            record.add_state_internal_engine_command_set(command, self._tick_time,
                                                         self._tick_number, self.tags_as_readonly())
            command.tick()
            if command.has_failed():
                record.add_state_failed(self._tick_time, self._tick_number, self.tags_as_readonly())
                cmds_done.add(cmd_request)
            elif command.is_finalized():
                record.add_state_completed(self._tick_time, self._tick_number, self.tags_as_readonly())
                cmds_done.add(cmd_request)
        else:
            logger.error(f"Runtime record is None for command {cmd_request}, this should not occur")

    def set_run_id(self) -> str:
        """ Creates a new run_id, sets the Run Id tag to it and returns it. """
        run_id = str(uuid.uuid4())
        self._system_tags[SystemTagName.RUN_ID].set_value(run_id, self._tick_time)
        return run_id

    def clear_run_id(self):
        self._system_tags[SystemTagName.RUN_ID].set_value(None, self._tick_time)

    def _stop_interpreter(self):
        self._interpreter.stop()
        self._interpreter = PInterpreter(self._method.get_program(), self)

    def _cancel_uod_commands(self):
        logger.debug("Cancelling uod commands")
        cmds_to_cancel: list[UodCommand] = []
        for name, command in self.uod.command_instances.items():
            if command.is_cancelled() or command.is_execution_complete() or command.is_finalized():
                logger.debug(f"Skipping command '{name}' that is no longer running")
            else:
                cmds_to_cancel.append(command)

        # call outside the loop because cancel modifies the collection
        for command in cmds_to_cancel:
            command.cancel()

    def _finalize_uod_commands(self):
        logger.debug("Cancelling uod commands")
        cmds_to_finalize: list[UodCommand] = []
        for command in self.uod.command_instances.values():
            if not command.is_finalized():
                cmds_to_finalize.append(command)

        # call outside the loop because finalize modifies the collection
        for command in cmds_to_finalize:
            command.finalize()

    def _execute_uod_command(self, cmd_request: CommandRequest, cmds_done: Set[CommandRequest]):
        cmd_name = cmd_request.name
        assert self.uod.has_command_name(cmd_name), f"Expected Uod to have command named '{cmd_name}'"
        assert cmd_request.exec_id is not None, f"Expected uod command request '{cmd_name}' to have exec_id"

        if not self.uod.hwl.is_connected:
            raise EngineError(
                f"The hardware is disconnected. The command '{cmd_name}' was not allowed to start.",
                "same")

        cancel_this = False

        # cancel any pending cancels (per user request)
        for c in self.cmd_executing:
            if c.command_exec_id in self._cancel_command_exec_ids:
                cmds_done.add(c)
                self._cancel_command_exec_ids.remove(c.command_exec_id)
                if cmd_request.name == c.name:
                    cancel_this = True
                cmd_record = self.runtimeinfo.get_command_and_record(c.command_exec_id)
                if cmd_record is not None:
                    command, c_record = cmd_record
                    command.cancel()
                    command.finalize()
                    c_record.add_command_state_cancelled(
                        c.command_exec_id, self._tick_time, self._tick_number, self.tags_as_readonly())
                    logger.info(f"Running command {c.name} cancelled per user request")
                else:
                    logger.error(f"Cannot cancel command {c}. No runtime record found with {c.exec_id=}" +
                                 f" and {c.command_exec_id=}")

        # cancel any existing instance with same name
        for c in [_c for _c in self.cmd_executing if _c not in cmds_done]:
            if c.name == cmd_name and not c == cmd_request:
                cmds_done.add(c)
                assert c.command_exec_id is not None, f"command_exec_id should be set for command '{cmd_name}'"
                cmd_record = self.runtimeinfo.get_command_and_record(c.command_exec_id)
                assert cmd_record is not None
                command, c_record = cmd_record
                command.cancel()
                command.finalize()
                c_record.add_command_state_cancelled(
                    c.command_exec_id, self._tick_time, self._tick_number,
                    self.tags_as_readonly())
                logger.debug(f"Running command {c.name} cancelled because another was started")

        # cancel any overlapping instance
        for c in [_c for _c in self.cmd_executing if _c not in cmds_done]:
            if not c == cmd_request:
                for overlap_list in self.uod.overlapping_command_names_lists:
                    if c.name in overlap_list and cmd_name in overlap_list:
                        cmds_done.add(c)
                        assert c.command_exec_id is not None, f"command_exec_id should be set for command '{c.name}'"
                        cmd_record = self.runtimeinfo.get_command_and_record(c.command_exec_id)
                        assert cmd_record is not None
                        command, c_record = cmd_record
                        command.cancel()
                        command.finalize()
                        c_record.add_command_state_cancelled(
                            c.command_exec_id, self._tick_time, self._tick_number,
                            self.tags_as_readonly())
                        logger.info(
                            f"Running command {c.name} cancelled because overlapping command " +
                            f"'{cmd_name}' was started")
                        break

        if cancel_this:
            # don't start command again that was just cancelled
            return

        record = self.interpreter.runtimeinfo.get_exec_record(cmd_request.exec_id)
        if record is None:
            logger.error(f"Failed to get record for command {cmd_request}")
            return

        # create or get command instance
        if not self.uod.has_command_instance(cmd_name):
            uod_command = self.uod.create_command(cmd_name)
            cmd_request.command_exec_id = record.add_state_uod_command_set(
                uod_command,
                self._tick_time,
                self._tick_number,
                self.tags_as_readonly())
        else:
            uod_command = self.uod.get_command(cmd_name)

        assert uod_command is not None, f"Failed to get uod_command for command '{cmd_name}'"

        logger.debug(f"Parsing arguments '{cmd_request.unparsed_args}' for uod command {cmd_name}")
        parsed_args = uod_command.parse_args(cmd_request.unparsed_args or "")

        if parsed_args is None:
            logger.error(f"Invalid argument string: '{cmd_request.unparsed_args}' for command '{cmd_name}'")
            cmds_done.add(cmd_request)
            raise ValueError(f"Invalid arguments for command '{cmd_name}'")

        # execute command state flow
        try:
            logger.debug(
                f"Executing uod command: '{cmd_request.name}' with parsed args '{parsed_args}', " +
                f"iteration {uod_command._exec_iterations}")
            if uod_command.is_cancelled():
                if not uod_command.is_finalized():
                    cmds_done.add(cmd_request)
                    uod_command.finalize()

            if not uod_command.is_initialized():
                uod_command.initialize()
                logger.debug(f"Command {cmd_request.name} initialized")

            if not uod_command.is_execution_started():
                assert cmd_request.command_exec_id is not None
                record.add_command_state_started(
                    cmd_request.command_exec_id,
                    self._tick_time, self._tick_number,
                    self.tags_as_readonly())
                uod_command.execute(parsed_args)
                logger.debug(f"Command {cmd_request.name} executed first iteration {uod_command._exec_iterations}")
            elif not uod_command.is_execution_complete():
                uod_command.execute(parsed_args)
                logger.debug(f"Command {cmd_request.name} executed another iteration {uod_command._exec_iterations}")

            if uod_command.is_execution_complete() and not uod_command.is_finalized():
                assert cmd_request.command_exec_id is not None
                record.add_command_state_completed(
                    cmd_request.command_exec_id,
                    self._tick_time, self._tick_number,
                    self.tags_as_readonly())
                cmds_done.add(cmd_request)
                uod_command.finalize()
                logger.debug(f"Command {cmd_request.name} finalized")

        except Exception as ex:
            if cmd_request in self.cmd_executing:
                cmds_done.add(cmd_request)
            assert cmd_request.command_exec_id is not None
            record.add_command_state_failed(
                cmd_request.command_exec_id,
                self._tick_time, self._tick_number,
                self.tags_as_readonly())
            cmd_record = self.runtimeinfo.get_command_and_record(cmd_request.command_exec_id)
            if cmd_record is not None:
                assert cmd_record[1].exec_id == record.exec_id
                command = cmd_record[0]
                if command is not None and not command.is_cancelled():
                    command.cancel()
                    logger.info(f"Cleaned up failed command {cmd_name}")

            logger.error(
                f"Uod command execution failed. Command: '{cmd_request.name}', " +
                f"argument string: '{cmd_request.unparsed_args}'", exc_info=True)
            raise ex

    def _apply_safe_state(self) -> TagValueCollection:
        current_values: List[TagValue] = []

        # TODO we should probably only consider uod tags here. would system tags ever have a safe value?
        for t in self._iter_all_tags():
            if t.direction == TagDirection.Output:
                safe_value = t.safe_value
                if not isinstance(safe_value, Unset):
                    current_values.append(t.as_readonly())
                    t.set_value(safe_value, self._tick_time)

        return TagValueCollection(current_values)

    def _apply_state(self, state: TagValueCollection):
        for t in self._iter_all_tags():
            if state.has(t.name):
                tag_value = state.get(t.name)
                t.set_value(tag_value.value, self._tick_time)

    def notify_all_tags(self):
        """ Collect tag updates from all tags, even unmodified tags """
        for tag in self._iter_all_tags():
            if tag.tick_time is None:
                logger.warning(f'Setting a tick time on {tag.name} tag missing it in notify_initial_tags()')
                tag.tick_time = self._tick_time
            self.tag_updates.put(tag)

    def notify_tag_updates(self):
        """ Collect tag updates from tags modified since last cycle """
        # pick up changes from listeners and queue them up
        for tag_name in self._system_listener.changes:
            tag = self._system_tags[tag_name]
            self.tag_updates.put(tag)
        self._system_listener.clear_changes()

        for tag_name in self._uod_listener.changes:
            tag = self.uod.tags[tag_name]
            self.tag_updates.put(tag)
        self._uod_listener.clear_changes()

    def set_error_state(self, exception: Exception):
        logger.info("Engine Paused because of error")
        self._system_tags[SystemTagName.METHOD_STATUS].set_value(MethodStatusEnum.ERROR, self._tick_time)
        self._system_tags[SystemTagName.SYSTEM_STATE].set_value(SystemStateEnum.Paused, self._tick_time)
        self._last_error = exception
        self._runstate_paused = True

    def has_error_state(self) -> bool:
        method_status = self._system_tags[SystemTagName.METHOD_STATUS]
        return method_status.get_value() == MethodStatusEnum.ERROR

    def get_error_state_exception(self) -> Exception | None:
        return self._last_error

    # TODO remove
    def parse_pcode(self, pcode: str) -> PProgram:
        p = PGrammar()
        p.parse(pcode)
        return p.build_model()

    def write_process_image(self):
        if not self._runstate_started:
            return

        hwl = self.uod.hwl
        register_values = []
        registers = [r for r in hwl.registers.values() if RegisterDirection.Write in r.direction]
        for r in registers:
            tag_value = self.uod.tags[r.name].get_value()
            register_value = tag_value if "from_tag" not in r.options \
                else r.options["from_tag"](tag_value)
            register_values.append(register_value)
        try:
            # possibly add guard for ConnectionState=Disconnected to avoid write errors
            # flooding the log
            hwl.write_batch(register_values, registers)
        except HardwareLayerException as ex:
            if not self.has_error_state():
                logger.error("Hardware write_batch error", exc_info=True)
                self.set_error_state(ex)

    def schedule_execution(self, name: str, exec_id: UUID | None = None, **kvargs):
        """ Execute named command from interpreter """
        if EngineCommandEnum.has_value(name) or self.uod.has_command_name(name):
            request = CommandRequest.from_interpreter(name, exec_id, **kvargs)
            self.cmd_queue.put_nowait(request)
        else:
            raise EngineError(
                f"Invalid command type scheduled: '{name}'",
                f"Unknown command: '{name}'"
            )

    def schedule_execution_user(self, name: str, args: str | None = None):
        """ Execute named command from user """
        # TODO args format needs to be specified in more detail. Its intended usage is
        # to contain argument values added to tag buttons. But in that case why not just use pcode?
        if EngineCommandEnum.has_value(name) or self.uod.has_command_name(name):
            if args is not None:
                raise NotImplementedError("User arguments format not implemented - command name: " + name)
            request = CommandRequest.from_user(name)
            self.cmd_queue.put_nowait(request)
        else:
            logger.error(f"Invalid command type scheduled: '{name}'")
            frontend_logger.error(f"Unknown command: '{name}'")
            raise Exception(f"Invalid command type scheduled: '{name}'")

    def inject_code(self, pcode: str):
        """ Inject a code snippet to run in the current scope of the current program """
        # TODO review signal Method change
        # TODO perform this change via _method
        logger.warning("TODO set updated method content and signal Method change")
        try:
            injected_program = self.parse_pcode(pcode)
            self.interpreter.inject_node(injected_program)
            logger.info("Injected code successful")
        except Exception as ex:
            logger.info("Injected code parse error: " + str(ex))
            self.set_error_state(ex)
            raise

    # code manipulation api
    def set_method(self, method: Mdl.Method):
        """ Set new method. This will replace the current method and invoke the on_method_init callback.

        On error, sets error_state and re-raises the error.
        """
        try:
            self._method.set_method(method)
            logger.debug(f"New method set with {len(method.lines)} lines")
        except Exception as ex:
            logger.error("Failed to set method", exc_info=True)
            self.set_error_state(ex)
            raise

    def cancel_instruction(self, exec_id: UUID):
        # try exec_id as a record exec_id
        record = self.runtimeinfo.get_exec_record(exec_id=exec_id)
        if record is not None:
            logger.info(f"Cancelling instruction {exec_id=}")
            record.node.cancel()
        else:
            # try exec_id as a command_exec_id
            result = self.runtimeinfo.get_command_and_record(command_exec_id=exec_id)
            if result is not None:
                _, record = result
                logger.info(f"Schedule cancellation of uod command {exec_id=}")
                self._cancel_command_exec_ids.add(exec_id)
                record.node.cancel()  # also need to mark the node as cancelled to update the runlog

        if record is None:
            logger.error(f"Cannot cancel instruction {exec_id=}, no runtime record exec_id or command_exec_id found")

    def force_instruction(self, exec_id: UUID):
        record = self.runtimeinfo.get_exec_record(exec_id=exec_id)
        if record is not None:
            logger.info(f"Forcing instruction with exec_id {exec_id=}")
            record.node.force()
        else:
            result = self.runtimeinfo.get_command_and_record(command_exec_id=exec_id)
            if result is not None:
                command, record = result
                logger.info(f"Forcing instruction with command_exec_id {exec_id=}")
                command.force()
                record.node.force()  # also need to mark the node as cancelled to update the runlog

        if record is None:
            logger.error(f"Cannot force instruction {exec_id=}, no runtime record found")


    def calculate_method_state(self) -> Mdl.MethodState:
        return self._method.calculate_method_state(self.interpreter.runtimeinfo)



import logging
from typing import Callable
from uuid import UUID
from openpectus.lang.exec.runlog import RuntimeInfo
from openpectus.lang.grammar.pgrammar import PGrammar
from openpectus.lang.model.pprogram import PNode, PProgram
import openpectus.protocol.models as Mdl

logger = logging.getLogger(__name__)

def parse_pcode(pcode: str) -> PProgram:
    p = PGrammar()
    p.parse(pcode)
    return p.build_model()


class MethodModel():
    """ Manages the mutable method state and its changes.

    Method state is mutated by engine as the program runs, e.g. executed lines are updated
    Method state is mutated by aggregator, e.g. by the user changing program code lines on injecting code
    """
    def __init__(self,
                 on_method_init: Callable[[], None] | None = None,
                 on_method_error: Callable[[Exception], None] | None = None) -> None:

        self._method = Mdl.Method.empty()
        self._method_state: Mdl.MethodState = Mdl.MethodState.empty()
        self._program: PProgram = PProgram()
        self._pcode: str = ""
        self._map_line_num_to_line_id: dict[int, str] = {}
        self._map_line_id_to_line_num: dict[str, int] = {}
        self._map_line_num_to_node_id: dict[int, UUID] = {}
        self._map_node_id_to_line_num: dict[UUID, int] = {}
        self._method_init_callback: Callable[[], None] | None = on_method_init
        self._method_error_callback: Callable[[Exception], None] | None = on_method_error

    def set_method(self, method: Mdl.Method):
        # initialize new method
        self._method = method
        self._pcode = '\n'.join(line.content for line in method.lines)
        try:
            self._program = parse_pcode(pcode=self._pcode)
        except Exception as ex:
            logger.error(f"Parse error, pcode:\n{self._pcode}", exc_info=True)
            if self._method_error_callback is not None:
                self._method_error_callback(ex)

            # TODO: Figure out whether we need to raise here - how is error callback used
            raise

        # fill mapping between line_num and line_id
        line_num: int = 1
        for line in method.lines:
            self._map_line_num_to_line_id[line_num] = line.id
            self._map_line_id_to_line_num[line.id] = line_num
            line_num += 1

        # fill mapping bewteen line_num and node_id
        def map_node(node: PNode):
            if node is not PProgram:
                assert node.line is not None
                self._map_line_num_to_node_id[node.line] = node.id
                self._map_node_id_to_line_num[node.id] = node.line
        self._program.iterate(map_node)

        if self._method_init_callback is not None:
            self._method_init_callback()

    def get_program(self) -> PProgram:
        return self._program

    def get_code_lines(self) -> list[Mdl.MethodLine]:
        return list([line for line in self._method.lines])

    def calculate_method_state(self, runtimeinfo: RuntimeInfo) -> Mdl.MethodState:
        method_state = Mdl.MethodState.empty()
        if self._pcode != "":
            for record in runtimeinfo.records:
                if not isinstance(record.node, PProgram):
                    assert record.node is not None, "node is None"
                    if record.node.line is not None:  # Note: injected nodes have no line numbers
                        line_num = record.node.line
                        line_id = self._map_line_num_to_line_id[line_num]
                        method_state.started_line_ids.append(line_id)
                        if record.visit_end_time != -1:
                            method_state.executed_line_ids.append(line_id)
        return method_state



import logging
import os
from argparse import ArgumentParser

from alembic import command
from alembic.config import Config
from openpectus import log_setup_colorlog
from openpectus.aggregator.aggregator_server import AggregatorServer
from openpectus import sentry, __version__, build_number

# - add lsp thingys
# - start (manage) lsp server instance for each client

# Start env in docker
# ../docker compose up --build

log_setup_colorlog(root_loglevel=logging.INFO)
logger = logging.getLogger("openpectus.aggregator.aggregator")
logger.setLevel(logging.INFO)
logging.getLogger("openpectus.protocol.aggregator_dispatcher").setLevel(logging.INFO)

def get_arg_parser():
    parser = ArgumentParser("Start Aggregator server")
    parser.add_argument("-host", "--host", required=False, default=AggregatorServer.default_host,
                        help=f"Host address to bind frontend and WebSocket to. Default: {AggregatorServer.default_host}")
    parser.add_argument("-p", "--port", required=False, type=int, default=AggregatorServer.default_port,
                        help=f"Host port to bind frontend and WebSocket to. Default: {AggregatorServer.default_port}")
    parser.add_argument("-fdd", "--frontend_dist_dir", required=False, default=AggregatorServer.default_frontend_dist_dir,
                        help=f"Frontend distribution directory. Default: {AggregatorServer.default_frontend_dist_dir}")
    parser.add_argument("-sev", "--sentry_event_level", required=False,
                        default=sentry.EVENT_LEVEL_DEFAULT, choices=sentry.EVENT_LEVEL_NAMES,
                        help=f"Minimum log level to send as sentry events. Default: '{sentry.EVENT_LEVEL_DEFAULT}'")
    parser.add_argument("-db", "--database", required=False, default=AggregatorServer.default_db_path,
                        help=f"Path to Sqlite3 database. Default: ./{AggregatorServer.default_db_filename}")
    parser.add_argument("-secret", "--secret", required=False, default=AggregatorServer.default_secret,
                        help="Engines must know this secret to connect to the aggregator")
    return parser


def main():
    args = get_arg_parser().parse_args()
    title = "Open Pectus Aggregator"
    logger.info(f"Starting {title} v. {__version__}, build: {build_number}")
    logger.info(f"Serving frontend at http://{args.host}:{args.port}")
    if os.getenv("SENTRY_DSN"):
        logger.info(f"Sentry is active with DSN={os.getenv('SENTRY_DSN')}")
    else:
        logger.info("Sentry is not active.")
    if os.getenv("ENABLE_AZURE_AUTHENTICATION", default="").lower() == "true":
        logger.info("Authentication is active.")
    else:
        logger.info("Authentication is not active.")
    sentry.init_aggregator(args.sentry_event_level)
    alembic_ini_file_path = os.path.join(os.path.dirname(__file__), "alembic.ini")
    alembic_config = Config(alembic_ini_file_path)
    alembic_config.set_main_option("sqlalchemy.url", f"sqlite:///{args.database}")
    command.upgrade(alembic_config, "head")
    AggregatorServer(title, args.host, args.port, args.frontend_dist_dir, args.database, args.secret).start()


if __name__ == "__main__":
    main()



"""Added aggregator_computer_name and aggregator_version to RecentRun

Revision ID: 5dbc834cf70e
Revises: 95bfa6af85ea
Create Date: 2024-02-06 14:30:02.409741

"""
from socket import gethostname
from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op
from openpectus import __version__

# revision identifiers, used by Alembic.
revision: str = '5dbc834cf70e'
down_revision: Union[str, None] = '95bfa6af85ea'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.add_column(sa.Column('aggregator_computer_name', sa.String(), nullable=False, server_default=gethostname()))
        batch_op.add_column(sa.Column('aggregator_version', sa.String(), nullable=False, server_default=__version__))
        batch_op.alter_column(column_name='aggregator_computer_name', server_default=None)
        batch_op.alter_column(column_name='aggregator_version', server_default=None)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('RecentRuns', schema=None) as batch_op:
        batch_op.drop_column('aggregator_version')
        batch_op.drop_column('aggregator_computer_name')

    # ### end Alembic commands ###



from enum import Flag, auto
import logging
from typing import Any, Sequence


logger = logging.getLogger(__name__)


class RegisterDirection(Flag):
    Read = auto()
    Write = auto()
    Both = Read | Write


class Register():
    """ Represents a hardware register or named value. """
    def __init__(self, name: str, direction: RegisterDirection, **options) -> None:
        self._name: str = name
        self._direction: RegisterDirection = direction
        # self._value: Any = None
        self._options: dict[str, Any] = options

    @property
    def name(self):
        return self._name

    @property
    def direction(self) -> RegisterDirection:
        return self._direction

    @property
    def options(self) -> dict[str, Any]:
        return self._options

    def __str__(self):
        return f"Register(name={self.name},direction={self.direction})"

    def __repr__(self):
        return str(self)


class HardwareLayerException(Exception):
    """ Raised when hardware connection issues occur. """

    def __init__(self, message: str, ex: Exception | None = None) -> None:
        self.message = message
        self.ex = ex

    def __str__(self):
        return self.message


class HardwareLayerBase():
    """ Base class for hardware layer implementations. """
    def __init__(self) -> None:
        self._registers: dict[str, Register] = {}
        self._is_connected: bool = False

    def __enter__(self):
        """ Allow use as context manager. """
        self.connect()

    def __exit__(self, type, value, traceback):
        self.disconnect()

    def tick(self):
        """ Invoked on each tick by engine. """
        pass

    def __str__(self) -> str:
        return f"HardwareLayer(type={type(self).__name__},registers={','.join(self.registers)})"

    @property
    def registers(self) -> dict[str, Register]:
        return self._registers

    @property
    def is_connected(self) -> bool:
        """ Returns a value indicating whether there is an active connection to the hardware. Virtual property. """
        return self._is_connected

    def read(self, r: Register) -> Any:
        """ Read single register value. Abstract method. """
        if RegisterDirection.Read not in r.direction:
            raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        raise NotImplementedError

    def read_batch(self, registers: Sequence[Register]) -> list[Any]:
        """ Read batch of register values. Override to provide efficient implementation. Virtual method. """
        for r in registers:
            if RegisterDirection.Read not in r.direction:
                raise HardwareLayerException(f"Attempt to read unreadable register {r}.")
        values = []
        for r in registers:
            values.append(self.read(r))
        return values

    def write(self, value: Any, r: Register) -> None:
        """ Write single register value. Abstract method. """
        if RegisterDirection.Write not in r.direction:
            raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        raise NotImplementedError

    def write_batch(self, values: Sequence[Any], registers: Sequence[Register]):
        """ Write batch of register values. Override to provide efficient implementation. Virtual method. """
        for r in registers:
            if RegisterDirection.Write not in r.direction:
                raise HardwareLayerException(f"Attempt to write unwritable register {r}.")
        for v, r in zip(values, registers):
            self.write(v, r)

    def connect(self):
        """ Connect to hardware. Throw HardwareLayerException on error. Virtual method.
        Implementations must call base method on completed connect.
        """
        self._is_connected = True

    def disconnect(self):
        """ Disconnect from hardware. Throw HardwareLayerException on error. Virtual method.
        Implementations must call base method on completed disconnect.
        """
        self._is_connected = False

    def reconnect(self):
        """ Perform a reconnect. The default implementation just disconnects (ignoring any error)
        and connects. Override to perform a more advanced reconnect, like using a thread.
        """
        logger.info("Reconnecting")
        try:
            logger.info("Reconnect - disconnect")
            self.disconnect()
        except Exception:
            logger.error("Reconnect - disconnect error", exc_info=True)
            # we continue - disconnect should not fail but even if it does,
            # connect might still work

        try:
            self.connect()
            logger.info("Reconnect complete")
        except Exception:
            logger.error("Reconnect - connect error", exc_info=True)
            raise HardwareLayerException("Reconnect attempt failed")


    def validate_offline(self):
        """ Perform checks that verify that the registers definition is valid. Raise on validation error. Virtual method. """
        pass

    def validate_online(self):
        """ Perform checks that verify online hardware access. Raise on validation error. Virtual method.

        The default implementation checks that all Read registers can be read. Override to add additional
        checks.
        """
        for r in self.registers.values():
            if RegisterDirection.Read in r.direction:
                _ = self.read(r)

    def show_online_register_details(self):
        """ Displays online register details to use in uod definition. """
        print("----- Register details -----")
        for r in self.registers.values():
            print(r)
            if RegisterDirection.Read in r.direction:
                try:
                    val = self.read(r)
                    print("Read successful")
                    print(f"Raw value read: '{val}', raw value type: {type(val)}")
                    if "to_tag" in r.options:
                        print("Register has 'to_tag' option defined")
                        try:
                            val = r.options["to_tag"](val)
                            print(f"'to_tag' value: '{val}', 'to_tag' value type: {type(val)}")
                        except Exception as ex:
                            print(f"Exception was raised trying to evaluate the to_tag option function: {ex}")
                except Exception:
                    logger.debug(f"Read of register '{r.name}' failed")
                    logger.error("Read exception", exc_info=True)
                print("")
        print("----------")


class NullHardware(HardwareLayerBase):
    """ Represents no hardware. Used by tests. """
    def __init__(self) -> None:
        super().__init__()
        self._is_connected = True

    def read(self, r: Register) -> Any:
        return None

    def write(self, value: Any, r: Register):
        pass

    # override calls we don't know to not fail
    def __getattribute__(self, name: str) -> Any:
        def null_call():
            pass

        if not name.startswith("_") and not hasattr(super(), name):
            # print("name", name)
            return null_call
        return super().__getattribute__(name)



import unittest

from openpectus.lang.exec.analyzer import (
    SemanticCheckAnalyzer,
    UnreachableCodeCheckAnalyzer,
    InfiniteBlockCheckAnalyzer,
    ConditionCheckAnalyzer,
    CommandCheckAnalyzer,
)
from openpectus.lang.exec.commands import CommandCollection, Command
from openpectus.lang.exec.tags import TagCollection, Tag
from openpectus.test.engine.utility_methods import build_program


# TODO present PErrorInstructions as errors
# TODO present PErrors as errors

class CommandCheckAnalyzerTest(unittest.TestCase):
    def test_command_defined(self):
        program = build_program("""
Foo
""")
        cmds = CommandCollection().with_cmd(Command("Foo"))
        analyzer = CommandCheckAnalyzer(cmds)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_command_undefined(self):
        program = build_program("""
Foo
""")
        cmds = CommandCollection()
        analyzer = CommandCheckAnalyzer(cmds)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))

    def test_command_arg_valid(self):
        program = build_program("""
Foo: bar
""")
        cmds = CommandCollection().with_cmd(Command("Foo"))
        analyzer = CommandCheckAnalyzer(cmds)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_command_arg_invalid(self):
        def is_valid_cmd(cmd_name):
            return cmd_name != 'bar'

        program = build_program("""
Foo: bar
""")
        cmds = CommandCollection().with_cmd(Command("Foo", is_valid_cmd))
        analyzer = CommandCheckAnalyzer(cmds)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))


class ConditionCheckAnalyzerTest(unittest.TestCase):
    def test_tag_name_defined(self):
        program = build_program("""
Watch: A > 2
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_tag_name_undefined(self):
        program = build_program("""
Watch: A > 2
    Mark: a
""")
        tags = TagCollection()
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))
        self.assertEqual("UndefinedTag", analyzer.items[0].id)

    def test_tag_unit_valid(self):
        program = build_program("""
Watch: A > 2
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_tag_unit_unexpected(self):
        program = build_program("""
Watch: A > 2 mL
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))
        self.assertEqual("UnexpectedUnit", analyzer.items[0].id)

    def test_tag_unit_missing(self):
        program = build_program("""
Watch: A > 2
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A", unit="mL"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))
        self.assertEqual("MissingUnit", analyzer.items[0].id)

    def test_tag_units_incompatible(self):
        program = build_program("""
Watch: A > 2 mL
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A", unit="s"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))
        self.assertEqual("IncompatibleUnits", analyzer.items[0].id)

    def test_tag_units_equal(self):
        program = build_program("""
Watch: A > 2 L
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A", unit="L"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_tag_units_compatible(self):
        program = build_program("""
Watch: A > 2 mL
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A", unit="L"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))

    def test_tag_unit_CV(self):
        # need to test CV comparison, CV being a custom unit unknown by pint
        program = build_program("""
Watch: A > 2 CV
    Mark: a
""")
        tags = TagCollection().with_tag(Tag("A", unit="CV"))
        analyzer = ConditionCheckAnalyzer(tags)
        analyzer.visit(program)
        self.assertEqual(0, len(analyzer.items))


    @unittest.skip("TODO")
    def test_tag_categorized_valid(self):
        pass

    @unittest.skip("TODO")
    def test_tag_categorized_invalid(self):
        pass


class UnreachableCodeCheckAnalyzerTest(unittest.TestCase):
    def test_UnreachableCodeVisitor(self):
        program = build_program("""
Block: A
    Mark: a
    End block
    Mark: b
Mark: c
""")
        analyzer = UnreachableCodeCheckAnalyzer()
        analyzer.visit(program)
        self.assertEqual(1, len(analyzer.items))


class InfiniteBlockCheckAnalyzerTest(unittest.TestCase):
    def test_missing_global_end(self):
        program = build_program("""
Block: A
    Mark: a
Mark: c
""")
        sa = InfiniteBlockCheckAnalyzer()
        sa.visit(program)
        self.assertEqual(1, len(sa.items))

    def test_has_global_end(self):
        program = build_program("""
Watch: Run counter > 0
    End block

Block: A
    Mark: a
Mark: c
""")
        sa = InfiniteBlockCheckAnalyzer()
        sa.visit(program)
        self.assertEqual(0, len(sa.items))

    def test_has_local_end(self):
        program = build_program("""
Block: A
    Mark: a
    Watch: Run counter > 0
        End block
Mark: c
""")
        sa = InfiniteBlockCheckAnalyzer()
        sa.visit(program)
        self.assertEqual(0, len(sa.items))

    def test_has_no_local_end(self):
        program = build_program("""
Block: A
    Mark: a
    Watch: Run counter > 0
        End block
Mark: c
Block: B
    Mark: b
""")
        sa = InfiniteBlockCheckAnalyzer()
        sa.visit(program)
        self.assertEqual(1, len(sa.items))


class SemanticCheckAnalyzerTest(unittest.TestCase):
    def test_basic(self):
        program = build_program("""
Mark: a
Mark: b
Mark: c
""")
        analyzer = SemanticCheckAnalyzer(TagCollection(), CommandCollection())
        analyzer.analyze(program)
        self.assertEqual(0, len(analyzer.errors))

    def test_warning_UnreachableCode(self):
        program = build_program("""
Block: A
    Mark: a
    End block
    Mark: b
Mark: c
""")
        analyzer = SemanticCheckAnalyzer(TagCollection(), CommandCollection())
        analyzer.analyze(program)
        xs = [e for e in analyzer.items if e.id == 'UnreachableCode']
        self.assertEqual(1, len(xs))



from typing import Sequence

import openpectus.protocol.messages as Msg
import openpectus.protocol.models as Mdl


# Note: These are the messages sent by EngineDispatcher and handled by AggregatorMessageHandlers.

class EngineMessage(Msg.MessageBase):
    sequence_number: int = -1
    engine_id: str = ''

    @property
    def ident(self) -> str:
        return f"seq: {self.sequence_number: >3}, type: {type(self).__name__}"


class EngineControlMessage(EngineMessage):
    pass


class EngineDataMessage(EngineMessage):
    pass


class PingMsg(EngineControlMessage):
    pass


class RegisterEngineMsg(Msg.MessageBase):
    """ Doesn't extend EngineMessage, because we don't have the engine_id yet """
    computer_name: str
    uod_name: str
    uod_author_name: str
    uod_author_email: str
    uod_filename: str
    location: str
    engine_version: str
    secret: str = ""
    # uod file hash, file change date

    sequence_number: int = -2

    @property
    def ident(self) -> str:
        return f"seq: {self.sequence_number: >3}, type: {type(self).__name__}"

    def __str__(self) -> str:
        return (f'{self.__class__.__name__}(computer_name="{self.computer_name}", ' +
                f'uod_name="{self.uod_name}", location="{self.location}")')


class UodInfoMsg(EngineMessage):
    readings: list[Mdl.ReadingInfo]
    commands: list[Mdl.CommandInfo]
    plot_configuration: Mdl.PlotConfiguration
    hardware_str: str
    required_roles: set[str]
    data_log_interval_seconds: float


class TagsUpdatedMsg(EngineMessage):
    tags: list[Mdl.TagValue] = []

# TODO: One possible way to supprt buffered messages. This would make it easier to persist
# the values for plot_log but maybe harder to insert in the realtime plot
class TagsBufferedMsg(EngineMessage):
    tags: list[Mdl.TagValue] = []
    tick_time: float


class RunLogMsg(EngineMessage):
    id: str
    run_id: str
    runlog: Mdl.RunLog


class ErrorLogMsg(EngineMessage):
    log: Mdl.ErrorLog


class MethodMsg(EngineMessage):
    method: Mdl.Method


class MethodStateMsg(EngineMessage):
    method_state: Mdl.MethodState


class ControlStateMsg(EngineMessage):
    control_state: Mdl.ControlState


class RunStartedMsg(EngineMessage):
    run_id: str
    started_tick: float


class RunStoppedMsg(EngineMessage):
    run_id: str
    runlog: Mdl.RunLog
    method: Mdl.Method


def print_sequence_range(messages: Sequence[EngineMessage]) -> str:
    """ print message sequence numbers in concise format"""
    if len(messages) == 0:
        return ""
    msgs_sorted: list[EngineMessage] = sorted(messages, key=lambda x: x.sequence_number)
    ranges = []
    range_start: int | None = None

    prev: EngineMessage | None = None
    for i, msg in enumerate(msgs_sorted):
        if i == 0:
            prev = msg
            range_start = msg.sequence_number
            continue
        if msg.sequence_number != prev.sequence_number + 1:  # type: ignore
            ranges.append((range_start, prev.sequence_number))  # type: ignore
            range_start = msg.sequence_number
        prev = msg

    if range_start is not None and range_start <= msgs_sorted[-1].sequence_number:
        ranges.append((range_start, msgs_sorted[-1].sequence_number))

    out = []
    for range in ranges:
        if range[0] == range[1]:
            out.append(str(range[0]))
        else:
            out.append(f"{range[0]}-{range[1]}")
    return ",".join(out)



import datetime
import logging
import time
import unittest
from typing import Any
from openpectus.engine.models import EngineCommandEnum
from openpectus.lang.exec.tags_impl import ReadingTag, SelectTag
from openpectus.engine.hardware import RegisterDirection

import pint
from openpectus.lang.exec.tags import SystemTagName, Tag, TagDirection, format_time_as_clock
from openpectus.lang.exec.uod import (
    UnitOperationDefinitionBase,
    UodCommand,
    UodBuilder,
    RegexNumber,
)
from openpectus.test.engine.utility_methods import (
    EngineTestRunner,
    configure_test_logger, set_engine_debug_logging, set_interpreter_debug_logging
)


configure_test_logger()
set_engine_debug_logging()
set_interpreter_debug_logging()
logging.getLogger("openpectus.lang.exec.runlog").setLevel(logging.DEBUG)

# pint takes forever to initialize - long enough
# to throw off timing of the first instruction.
# so we initialize it first
_ = pint.Quantity("0 s")


delta = 0.1

def create_test_uod() -> UnitOperationDefinitionBase:
    def reset(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        if count == 0:
            cmd.context.tags.get("Reset").set_value("Reset", time.time())
        elif count == 4:
            cmd.context.tags.get("Reset").set_value("N/A", time.time())
            cmd.set_complete()

    def cmd_with_args(cmd: UodCommand, **kvargs) -> None:
        print("arguments: " + str(kvargs))
        cmd.set_complete()

    def cmd_arg_parse(args: str) -> dict[str, Any] | None:
        if args is None or args == "" or args == "FAIL":
            return None
        result = {}
        for i, item in enumerate(args.split(",")):
            result[f"item{i}"] = item.strip()
        return result

    def cmd_regex(cmd: UodCommand, number: str, number_unit: str) -> None:
        cmd.context.tags["CmdWithRegex_Flowrate"].set_value_and_unit(float(number), number_unit, time.time())
        cmd.set_complete()

    def overlap_exec(cmd: UodCommand, **kvargs) -> None:
        count = cmd.get_iteration_count()
        if count >= 9:
            cmd.set_complete()

    builder = (
        UodBuilder()
        .with_instrument("TestUod")
        .with_author("Test Author", "test@openpectus.org")
        .with_filename(__file__)
        .with_hardware_none()
        .with_location("Test location")
        .with_hardware_register("FT01", RegisterDirection.Both, path="Objects;2:System;2:FT01")
        .with_hardware_register(
            "Reset",
            RegisterDirection.Both,
            path="Objects;2:System;2:RESET",
            from_tag=lambda x: 1 if x == "Reset" else 0,
            to_tag=lambda x: "Reset" if x == 1 else "N/A",
        )
        # Readings
        .with_tag(ReadingTag("FT01", "L/h"))
        .with_tag(SelectTag("Reset", value="N/A", unit=None, choices=["Reset", "N/A"]))
        .with_tag(Tag("Danger", value=True, unit=None, direction=TagDirection.Output, safe_value=False))
        .with_command(name="Reset", exec_fn=reset)
        .with_command(name="CmdWithArgs", exec_fn=cmd_with_args, arg_parse_fn=cmd_arg_parse)
        .with_command(name="overlap1", exec_fn=overlap_exec)
        .with_command(name="overlap2", exec_fn=overlap_exec)
        .with_command_overlap(['overlap1', 'overlap2'])
        .with_tag(ReadingTag("CmdWithRegex_Flowrate", "L/h"))
        .with_command_regex_arguments(
            name="CmdWithRegexArgs",
            arg_parse_regex=RegexNumber(units=['L/h', 'L/min', 'L/s']),
            exec_fn=cmd_regex)
    )
    uod = builder.build()
    uod.hwl.connect()
    return uod


class TestEngineTags(unittest.TestCase):
    # ---------- Lifetime -------------
    def test_tag_process_time(self):
        runner = EngineTestRunner(create_test_uod)
        with runner.run() as instance:
            e = instance.engine
            instance.start()
            instance.run_ticks(2)

            process_time = e.tags[SystemTagName.PROCESS_TIME]
            self.assertAlmostEqual(0.2, process_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.STOP)
            instance.run_ticks(10)
            self.assertAlmostEqual(0.3, process_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.START)
            instance.run_ticks(5)

            # is reset on system Start
            self.assertAlmostEqual(0.5, process_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.PAUSE)
            instance.run_ticks(5)

            # is paused while system is Paused
            self.assertAlmostEqual(0.6, process_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.UNPAUSE)
            instance.run_ticks(5)

            self.assertAlmostEqual(1.0, process_time.as_float(), delta=delta)

    def test_tag_run_time(self):
        runner = EngineTestRunner(create_test_uod)
        with runner.run() as instance:
            e = instance.engine
            instance.start()

            instance.run_ticks(2)

            run_time = e.tags[SystemTagName.RUN_TIME]
            self.assertAlmostEqual(0.2, run_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.STOP)
            instance.run_ticks(10)

            self.assertAlmostEqual(0.3, run_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.START)
            instance.run_ticks(5)

            # is reset on system Start
            self.assertAlmostEqual(0.5, run_time.as_float(), delta=delta)

            e.schedule_execution(EngineCommandEnum.PAUSE)
            instance.run_ticks(5)

            # not paused while system is Paused
            self.assertAlmostEqual(1.0, run_time.as_float(), delta=delta)

    def test_tag_block_time(self):
        p = """
Wait: 0.25s
Block: B1
    Wait: 0.25s
    End block
"""
        runner = EngineTestRunner(create_test_uod, p)
        with runner.run() as instance:
            e = instance.engine
            instance.start()

            instance.run_ticks(2)

            block_time = e.tags[SystemTagName.BLOCK_TIME]
            block = e.tags[SystemTagName.BLOCK]

            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.2, block_time.as_number(), delta=delta)

            instance.run_ticks(4)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.6, block_time.as_number(), delta=delta)

            instance.run_ticks(1)
            self.assertEqual("B1", block.get_value())
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)

            instance.run_ticks(6)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.1 + 0.6, block_time.as_number(), delta=delta)

    # @unittest.skip("Test is likely wrong. Should discuss nested block times")
    def test_tag_block_time_nested_blocks(self):
        p = """
Block: B1
    Wait: 0.15s
    Block: B2
        Wait: 0.05s
        End block
    Wait: 0.25s
    End block
"""
        runner = EngineTestRunner(create_test_uod, p)
        with runner.run() as instance:
            e = instance.engine
            instance.start()

            block_time = e.tags[SystemTagName.BLOCK_TIME]
            block = e.tags[SystemTagName.BLOCK]

            instance.run_ticks(2)

            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.2, block_time.as_number(), delta=delta)

            instance.run_until_event("block_start")
            self.assertEqual("B1", block.get_value())
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)
            instance.run_ticks(3)
            self.assertEqual("B1", block.get_value())
            self.assertAlmostEqual(0.4, block_time.as_number(), delta=delta)

            instance.run_until_event("block_start")
            self.assertEqual("B2", block.get_value())
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)

            instance.run_ticks(3)
            self.assertEqual("B2", block.get_value())
            self.assertAlmostEqual(0.4, block_time.as_number(), delta=delta)

            instance.run_until_event("block_end")
            # continue_engine(e, 1)
            self.assertEqual("B1", block.get_value())
            # TODO currently does not increment block B1 timer while B2 is active
            self.assertAlmostEqual(0.5, block_time.as_number(), delta=delta)  # 0.4 + 0.4 ???
            # self.assertAlmostEqual(0.5 + 0.4, block_time.as_number(), delta=delta)  # 0.4 + 0.4 ???
            # continue_engine(e, 4)
            # self.assertEqual("B1", block.get_value())
            # self.assertAlmostEqual(0.9, block_time.as_number(), delta=delta)

            # continue_engine(e, 1)
            # self.assertEqual(None, block.get_value())
            # self.assertAlmostEqual(0.2, block_time.as_number(), delta=delta)  # 0.9 + 0.2 ???

    def test_tag_block_time_restart(self):
        p = """
Wait: 0.25s
Block: B1
    Wait: 0.25s
    End block
"""
        runner = EngineTestRunner(create_test_uod, p)
        with runner.run() as instance:
            e = instance.engine
            instance.start()
            block_time = e.tags[SystemTagName.BLOCK_TIME]
            block = e.tags[SystemTagName.BLOCK]

            instance.run_ticks(2)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.2, block_time.as_number(), delta=delta)

            instance.run_ticks(4)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.6, block_time.as_number(), delta=delta)

            instance.run_ticks(1)
            self.assertEqual("B1", block.get_value())
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)

            instance.run_ticks(6)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.1 + 0.6, block_time.as_number(), delta=delta)

            e.schedule_execution(EngineCommandEnum.RESTART)
            instance.run_ticks(3)
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)  # is reset after restart

            # instance.run_until_condition(lambda: block.get_value() == "B1")
            instance.run_until_event("block_start")
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)  # nested timer is also reset

    def test_tag_block_time_stop_start(self):
        p = """
Wait: 0.25s
Block: B1
    Wait: 0.25s
    End block
"""
        runner = EngineTestRunner(create_test_uod, p)
        with runner.run() as instance:
            e = instance.engine
            instance.start()
            block_time = e.tags[SystemTagName.BLOCK_TIME]
            block = e.tags[SystemTagName.BLOCK]

            instance.run_ticks(2)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.2, block_time.as_number(), delta=delta)

            instance.run_ticks(4)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.6, block_time.as_number(), delta=delta)

            instance.run_ticks(1)
            self.assertEqual("B1", block.get_value())
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)

            instance.run_ticks(6)
            self.assertEqual(None, block.get_value())
            self.assertAlmostEqual(0.1 + 0.6, block_time.as_number(), delta=delta)

            e.schedule_execution(EngineCommandEnum.STOP)
            instance.run_ticks(3)

            e.schedule_execution(EngineCommandEnum.START)
            instance.run_ticks(1)
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)  # is reset after restart

            instance.run_until_event("block_start")
            self.assertAlmostEqual(0.1, block_time.as_number(), delta=delta)  # nested timer is also reset

    def test_tag_unit_conversion_handled_by_pint(self):
        p = """
CmdWithRegexArgs: 2 L/h
CmdWithRegexArgs: 2 L/min
CmdWithRegexArgs: 2 L/s
"""
        runner = EngineTestRunner(create_test_uod, p)
        with runner.run() as instance:
            e = instance.engine
            instance.start()
            flowrate = e.tags["CmdWithRegex_Flowrate"]  # [L/h]

            instance.run_ticks(1)
            self.assertEqual(0.0, flowrate.get_value())
            instance.run_ticks(2)
            self.assertEqual(2.0, flowrate.get_value())  # 2 L/h is 2 L/h
            instance.run_ticks(1)
            self.assertEqual(120.0, flowrate.get_value())  # 2 L/min is 120 L/h
            instance.run_ticks(1)
            self.assertEqual(7200.0, flowrate.get_value())  # 2 L/s is 7200 L/h



class TestFormatting(unittest.TestCase):

    def test_formatting(self):
        tag = Tag("foo")

        tag.format_fn = format_time_as_clock
        _date = datetime.datetime(
            year=2000, month=1, day=1, hour=6, minute=46, second=3, microsecond=111, tzinfo=datetime.UTC
        )
        tag.value = _date.timestamp()

        t = tag.as_readonly()

        self.assertEqual("06:46:03", t.value_formatted)
        self.assertIsNotNone(t.value_formatted)



import csv
import logging
from io import StringIO
from operator import concat
from typing import Iterable

from openpectus.aggregator.routers.dto import PlotLog, RecentRun, ProcessValueValueType

logger = logging.getLogger(__name__)


def generate_csv_string(plot_log: PlotLog, recent_run: RecentRun):
    csv_string = StringIO()
    csv_writer = csv.writer(csv_string)
    _write_metadata_rows(csv_writer, plot_log, recent_run)
    _write_header_row(csv_writer, plot_log)
    _write_data_rows(csv_writer, plot_log, _get_tick_times(plot_log))
    return csv_string


def _write_metadata_rows(csv_writer, plot_log: PlotLog, recent_run: RecentRun):
    csv_writer.writerow(['# Recent Run Id', recent_run.run_id])
    csv_writer.writerow(['# Engine Id', recent_run.engine_id])
    csv_writer.writerow(['# Engine Computer name', recent_run.engine_computer_name])
    csv_writer.writerow(['# Engine Version', recent_run.engine_version])
    csv_writer.writerow(['# Engine Hardware string', recent_run.engine_hardware_str])
    csv_writer.writerow(['# Uod author name', recent_run.uod_author_name])
    csv_writer.writerow(['# Uod author email', recent_run.uod_author_email])
    csv_writer.writerow(['# Uod file name', recent_run.uod_filename])
    csv_writer.writerow(['# Aggregator Computer name', recent_run.aggregator_computer_name])
    csv_writer.writerow(['# Aggregator Version', recent_run.aggregator_version])
    csv_writer.writerow(['# Starting Time (UTC)', recent_run.started_date])
    csv_writer.writerow(['# Ending Time (UTC)', recent_run.completed_date])
    csv_writer.writerow(concat(['# Contributors'], recent_run.contributors))
    csv_writer.writerow([])


def _get_tick_times(plot_log: PlotLog):
    list_of_all_tick_times = (value.tick_time for entry in plot_log.entries.values() for value in entry.values)
    unique_tick_times = list(set(list_of_all_tick_times))
    unique_tick_times.sort()
    return unique_tick_times


def _write_data_rows(csv_writer, plot_log: PlotLog, unique_tick_times: Iterable[float]):
    for tick_time in unique_tick_times:
        row: list[ProcessValueValueType] = []
        for entry in plot_log.entries.values():
            if len(entry.values) == 0:
                row.append(None)
                continue
            elif len(entry.values) >= 2 and tick_time >= entry.values[1].tick_time:
                entry.values.pop(0)
            row.append(entry.values[0].value)
        csv_writer.writerow(row)


def _write_header_row(csv_writer, plot_log: PlotLog):
    # write header row and ensure values are sorted by tick_time
    header_row: list[str] = []
    for entry in plot_log.entries.values():
        entry.values.sort(key=lambda e: e.tick_time)
        header_row.append(entry.name + (f' [{entry.value_unit}]' if entry.value_unit is not None else ''))
    csv_writer.writerow(header_row)



import os
from typing import Annotated, Any

import jwt
from openpectus.aggregator.data.models import RecentEngine, RecentRun
from openpectus.aggregator.models import EngineData
from fastapi import APIRouter, HTTPException, status
from fastapi.params import Header, Security
from openpectus.aggregator.routers.dto import AuthConfig

router = APIRouter(tags=["auth"])

# Much in this file is based on https://alukach.com/posts/fastapi-rs256-jwt/
# Authentication is possible via PKCE flow or via client secret for daemon
# applications.
# Remember to add optional claim "idtyp" on Azure App Registration -> Token configuration page.

use_auth = os.getenv("ENABLE_AZURE_AUTHENTICATION", default="").lower() == "true"
tenant_id = os.getenv("AZURE_DIRECTORY_TENANT_ID", default=None)
client_id = os.getenv("AZURE_APPLICATION_CLIENT_ID", default=None)
authority_url = f"https://login.microsoftonline.com/{tenant_id}/v2.0" if tenant_id else None
well_known_url = f"{authority_url}/.well-known/openid-configuration" if tenant_id else None
jwks_url_id_token = "https://login.microsoftonline.com/common/discovery/keys"
jwks_url_access_token = f"https://login.microsoftonline.com/{tenant_id}/discovery/v2.0/keys" if tenant_id else ""
access_token_issuer = f"https://sts.windows.net/{tenant_id}/" if tenant_id else ""

if use_auth:
    assert tenant_id is not None, "Assign Tenant ID to environment variable AZURE_DIRECTORY_TENANT_ID"
    assert client_id is not None, "Assign Client ID to environment variable AZURE_APPLICATION_CLIENT_ID"

@router.get("/config")
def get_config() -> AuthConfig:
    return AuthConfig(
        use_auth=use_auth,
        authority_url=authority_url,
        client_id=client_id,
        well_known_url=well_known_url
    )


# PKCE and client secret flows use different keys:
jwks_client_pkce = jwt.PyJWKClient(jwks_url_id_token)
jwks_client_secret = jwt.PyJWKClient(jwks_url_access_token)


def decode_token_or_fail(x_identity) -> dict[str, Any]:
    # Bearer tokens from PKCE flow result in an ID token.
    # Bearer tokens from secret flow result in access token.
    # Try to parse either one. Raise exception if neither
    # can be validated.
    try:
        id_token: dict[str, str | list[str]] = jwt.decode(
            x_identity,
            jwks_client_pkce.get_signing_key_from_jwt(x_identity).key,
            algorithms=["RS256"],
            audience=client_id,
            issuer=authority_url
        )
        return id_token
    except jwt.PyJWTError:
        pass

    try:
        access_token: dict[str, str | list[str]] = jwt.decode(
            x_identity,
            jwks_client_secret.get_signing_key_from_jwt(x_identity),
            algorithms=["RS256"],
            audience=client_id,
            issuer=access_token_issuer
        )
        return access_token
    except jwt.PyJWTError:
        pass
    # If we got this far, then we couldn't validate the token in either case
    raise HTTPException(
        status_code=status.HTTP_401_UNAUTHORIZED,
        detail="Could not validate token in X-Identity header",
        headers={"WWW-Authenticate": "Bearer"},
    )


def user_name(x_identity: Annotated[str, Header()] = "") -> str:
    if not use_auth:
        return "Anon"

    token = decode_token_or_fail(x_identity)
    # Return default name "Daemon" for authenticated applications
    if token.get("idtyp", "") == "app":
        return "Daemon"
    preferred_username = token.get("preferred_username", "")  # email address, field 'name' has full name
    if "@" in preferred_username:
        preferred_username = preferred_username[0: preferred_username.index("@")]
    return str(preferred_username).upper()


def user_roles(x_identity: Annotated[str, Header()] = "") -> set[str]:
    if not use_auth:
        return set()

    token = decode_token_or_fail(x_identity)
    roles = set(token.get("roles") or [])
    # Add default role "Daemon" for authenticated applications
    if token.get("idtyp", "") == "app":
        roles |= set(["Daemon"])
    return roles


UserRolesDependency = Security(user_roles)
UserRolesValue = Annotated[set[str], UserRolesDependency]
UserNameDependency = Security(user_name)
UserNameValue = Annotated[str, UserNameDependency]


def has_access(engine_or_run: EngineData | RecentEngine | RecentRun, user_roles: set[str]):
    required_roles = set(engine_or_run.required_roles)
    return len(required_roles) == 0 or len(required_roles & user_roles) > 0



import logging
import time
import unittest

import pint
from openpectus.engine.engine import Engine, EngineTiming
from openpectus.engine.models import EngineCommandEnum
from openpectus.lang.exec.analyzer import ConditionEnrichAnalyzer
from openpectus.lang.exec.clock import WallClock
from openpectus.lang.exec.pinterpreter import PInterpreter
from openpectus.lang.exec.tags import Tag, SystemTagName
from openpectus.lang.exec.timer import NullTimer
from openpectus.lang.exec.uod import UnitOperationDefinitionBase, UodBuilder, UodCommand
from openpectus.lang.grammar.pprogramformatter import print_parsed_program as print_program
from openpectus.lang.model.pprogram import PCondition, PNode, PProgram, PWatch
from openpectus.test.engine.utility_methods import (
    continue_engine, run_engine, build_program,
    configure_test_logger, set_engine_debug_logging, set_interpreter_debug_logging,
    print_runlog
)


configure_test_logger()
set_engine_debug_logging()
set_interpreter_debug_logging()


# pint takes forever to initialize - long enough
# to throw off timing of the first instruction.
# so we initialize it first
_ = pint.Quantity("0 sec")


def create_test_uod() -> UnitOperationDefinitionBase:
    def incr_counter(cmd: UodCommand, **kvargs):
        counter = cmd.context.tags["counter"]
        count = counter.as_number()
        count = count + 1
        counter.set_value(count, time.time())
        cmd.set_complete()

    return (
        UodBuilder()
        .with_instrument("TestUod")
        .with_author("Test Author", "test@openpectus.org")
        .with_filename(__file__)
        .with_hardware_none()
        .with_location("Test location")
        # Readings
        .with_tag(Tag(name="counter", value=0))
        .with_command(name="incr counter", exec_fn=incr_counter)
        .build()
    )


def create_engine(uod: UnitOperationDefinitionBase | None = None) -> Engine:
    if uod is None:
        uod = create_test_uod()
    e = Engine(uod, EngineTiming(WallClock(), NullTimer(), 0.1, 1.0))
    return e


def create_interpreter(
        program: PProgram,
        uod: UnitOperationDefinitionBase | None = None):
    # create interpreter without engine - for non-command programs

    if uod is None:
        uod = create_test_uod()

    engine = create_engine(uod)

    return PInterpreter(program, engine)


def run_interpreter(interpreter: PInterpreter, max_ticks: int = -1):
    print("Interpretation started")
    ticks = 0
    max_ticks = max_ticks
    interpreter.running = True

    while interpreter.running:
        ticks += 1
        if max_ticks != -1 and ticks > max_ticks:
            print(f"Stopping because max_ticks {max_ticks} was reached")
            interpreter.running = False
            return

        time.sleep(0.1)
        tick_time = time.time()
        interpreter.tick(tick_time, ticks)


class InterpreterTest(unittest.TestCase):

    def setUp(self):
        self.engine: Engine = create_engine()

    def tearDown(self):
        self.engine.cleanup()

    def test_sequential_marks(self):
        program = """
Mark: a
Mark: b
Mark: c
"""
        print_program(program)
        engine = self.engine
        run_engine(engine, program, 10)

        self.assertEqual(["a", "b", "c"], engine.interpreter.get_marks())

    def test_command_incr_counter(self):
        program = """
Mark: a
incr counter
"""
        print_program(program)
        engine = self.engine

        self.assertEqual(0, engine.uod.tags["counter"].as_number())

        run_engine(engine, program, 10)

        self.assertEqual(1, engine.uod.tags["counter"].as_number())
        self.assertEqual("a", engine.interpreter.get_marks()[0])

    def test_command_Increment_run_counter(self):
        # same as test_command_incr_counter() but using builtin system tag and command rather than uod variant.
        program = """
Mark: a
Increment run counter
"""
        print_program(program)
        engine = self.engine

        self.assertEqual(0, engine.tags[SystemTagName.RUN_COUNTER].as_number())

        run_engine(engine, program, 5)

        self.assertEqual(1, engine.tags[SystemTagName.RUN_COUNTER].as_number())
        self.assertEqual(["a"], engine.interpreter.get_marks())

    def test_command_Run_counter(self):
        program = "Run counter: 3"
        engine = self.engine

        self.assertEqual(0, engine.tags[SystemTagName.RUN_COUNTER].as_number())

        run_engine(engine, program, 5)

        self.assertEqual(3, engine.tags[SystemTagName.RUN_COUNTER].as_number())

    @unittest.skip("TODO")
    def test_condition_with_invalid_tag_fails(self):
        raise NotImplementedError()

    def test_watch_can_evaluate_tag(self):
        program = """
Mark: a
Watch: counter > 0
    Mark: b
Mark: c
incr counter
Watch: counter > 0
    Mark: d
"""
        engine = self.engine
        run_engine(engine, program, 15)

        self.assertEqual(["a", "c", "b", "d"], engine.interpreter.get_marks())
        # Note that first watch is also activated and its body executed
        # even though it is not activated when first run. This represents the
        # interrupt nature of watches and alarms. So the behavior is not like this:
        # self.assertEqual(["a", "c", "d"], i.get_marks())


# --- Conditions ---


    def test_evaluate(self):
        self.engine.interpreter.context.tags.add(Tag("X", value=9, unit="%"))

        parent = PNode(None)
        c = PWatch(parent)
        c.condition = PCondition("X > 10%")
        c.condition.lhs = "X"
        c.condition.op = ">"
        c.condition.rhs = "10%"
        # TODO get rid of enrich analyzers - they are really just in the way
        an = ConditionEnrichAnalyzer()
        an.enrich_condition(c)
        result = self.engine.interpreter._evaluate_condition(c)
        self.assertEqual(result, False)



    def test_watch_nested(self):
        program = """
Mark: a
Watch: counter > 0
    Mark: b
    Watch: counter > 1
        Mark: e
Mark: c
incr counter
incr counter
Watch: counter > 0
    Mark: d
"""
        engine = self.engine
        run_engine(engine, program, 15)

        # TODO abgiguous ...
        # self.assertEqual(["a", "c", "b", "e", "d"], engine.interpreter.get_marks())
        self.assertEqual(["a", "c", "b", "d", "e"], engine.interpreter.get_marks())

    def test_watch_restart(self):
        program = """
Mark: a
Watch: counter > 0
    Mark: b
Mark: c
incr counter
"""
        engine = self.engine
        run_engine(engine, program, 15)
        self.assertEqual(["a", "c", "b"], engine.interpreter.get_marks())

        engine.schedule_execution(EngineCommandEnum.STOP)
        continue_engine(engine, 5)

        # rerun program and assert "same" result
        engine.schedule_execution(EngineCommandEnum.START)
        continue_engine(engine, 15)
        self.assertEqual(["a", "c", "b"], engine.interpreter.get_marks())

    def test_watch_long_running_order(self):
        # specify order of highly overlapping instructions
        # between main program and a single interrupt handler
        program = """
Mark: a
Watch: Run Counter > -1
    Mark: a1
    Mark: a2
    Mark: a3
Mark: b
Mark: b1
Mark: b2
Mark: b3
"""
        engine = self.engine
        run_engine(engine, program, 10)

        self.assertEqual(["a", "b", "a1", "b1", "a2", "b2", "a3", "b3"], engine.interpreter.get_marks())

    @unittest.skip("Block in Watch not supported")
    def test_watch_block_long_running_block_time(self):
        # question for another test: is Block time defined for alarm without block? Yes

        # specify block time for block in interrupt handler

        program = """
Mark: a
Watch: Run counter > -1
    Block: A
        Mark: a1
        Watch: Block Time > 0.5 sec
            Mark: a2
        Watch: Block Time > 1.0 sec
            Mark: a3
        Watch: Block Time > 1.2 sec
            Mark: a32
        Watch: Block Time > 1.3 sec
            Mark: a33
        Watch: Block Time > 1.4 sec
            Mark: a34
        Watch: Block Time > 1.5 sec
            Mark: a4
            End block
Mark: b
"""
        print_program(program)
        engine = self.engine
        run_engine(engine, program, 15)

        self.assertEqual(["a", "b", "a1", "a2", "a3", "a4"], engine.interpreter.get_marks())

    def test_alarm_triggers_by_condition(self):
        program = """
Mark: a
Alarm: counter > 0
    Mark: b
Mark: c
incr counter
Mark: d
"""
        engine = self.engine
        run_engine(engine, program, 15)

        marks = engine.interpreter.get_marks()

        # most important
        self.assertEqual(["a", "c"], marks[:2])
        # also important but specific order is really an implementation detail
        self.assertTrue(marks[:4] == ["a", "c", "d", "b"] or marks[:4] == ["a", "c", "b", "d"])

    def test_alarm_can_retrigger(self):
        program = """
Mark: a
Alarm: counter > 0
    Mark: b
Mark: c
incr counter
Mark: d
Mark: e
Mark: f
"""
        engine = self.engine

        logger = logging.getLogger("openpectus.lang.exec.pinterpreter")
        logger.setLevel(logging.DEBUG)

        run_engine(engine, program, 15)

        print_runlog(engine)
        # print_runtime_records(engine)

        marks = engine.interpreter.get_marks()
        # most important
        self.assertEqual(["a", "c"], marks[:2])
        # least important - exact ordering is an implementation
        self.assertGreaterEqual(marks.count("b"), 2)

    def test_block(self):
        program = """
Block: A
    Mark: A1
    End block
    Mark: A2
Mark: A3
"""
        engine = self.engine
        run_engine(engine, program, 10)

        self.assertEqual(["A1", "A3"], engine.interpreter.get_marks())

    def test_block_nested(self):
        program = """
Block: A
    Mark: A1
    Block: B
        Mark: B1
        End block
        Mark: B2
    End block
    Mark: A2
Mark: A3
"""
        engine = self.engine
        run_engine(engine, program, 15)

        self.assertEqual(["A1", "B1", "A3"], engine.interpreter.get_marks())

    def test_block_unterminated(self):
        program = """
Block: A
    Mark: A1
    Mark: A2
Mark: A3
"""
        engine = self.engine
        run_engine(engine, program, 5)

        self.assertEqual(["A1", "A2"], engine.interpreter.get_marks())

    def test_block_time_watch(self):
        program = """
Block: A
    Mark: A1
    Watch: Block Time > 1 s
        End block
    Mark: A2
Mark: A3
"""
        engine = self.engine
        run_engine(engine, program, 20)

        self.assertEqual(["A1", "A2", "A3"], engine.interpreter.get_marks())


    def test_block_time_watch_complex(self):
        program = """
Block: A
    Mark: A1
    Watch: Block Time > 0.5 s
        Mark: A2
    Watch: Block Time > 1 s
        Mark: A3
        End block
    Mark: A4
Mark: A5
"""
        engine = self.engine
        run_engine(engine, program, 25)

        self.assertEqual(["A1", "A4", "A2", "A3", "A5"], engine.interpreter.get_marks())

    def test_block_end_block(self):
        program = """
Block: A
    Mark: A1
    End block
Mark: A2
"""
        engine = self.engine
        run_engine(engine, program, 10)

        self.assertEqual(["A1", "A2"], engine.interpreter.get_marks())

    def test_block_nested_end_block(self):
        program = """
Block: A
    Mark: A1
    Block: B
        Mark: B1
        End block
Mark: A2
"""
        engine = self.engine
        run_engine(engine, program, 15)

        self.assertEqual(["A1", "B1"], engine.interpreter.get_marks())

    def test_block_end_blocks(self):
        program = """
Block: A
    Mark: A1
    End blocks
Mark: A2
"""
        engine = self.engine
        run_engine(engine, program, 10)


        self.assertEqual(["A1", "A2"], engine.interpreter.get_marks())

    def test_block_nested_end_blocks(self):
        program = """
Block: A
    Mark: A1
    Block: B
        Mark: B1
        End blocks
Mark: A2
"""
        engine = self.engine
        run_engine(engine, program, 15)

        self.assertEqual(["A1", "B1", "A2"], engine.interpreter.get_marks())

    def test_base_command_sec(self):
        program = "Base: s"
        e = self.engine
        run_engine(e, program, 5)
        base_tag = e._system_tags[SystemTagName.BASE]
        self.assertEqual("s", base_tag.get_value())

    def test_base_command_min(self):
        program = "Base: min"
        e = self.engine
        run_engine(e, program, 5)
        base_tag = e._system_tags[SystemTagName.BASE]
        self.assertEqual("min", base_tag.get_value())

    def test_base_command_h(self):
        program = "Base: h"
        e = self.engine
        run_engine(e, program, 5)
        base_tag = e._system_tags[SystemTagName.BASE]
        self.assertEqual("h", base_tag.get_value())

    def test_base_default_is_set_to_minutes(self):
        e = self.engine
        base_tag = e._system_tags[SystemTagName.BASE]
        self.assertIsNotNone(base_tag)
        self.assertEqual(base_tag.unit, None)
        self.assertEqual(base_tag.value, "min")

    def test_base_can_set(self):
        e = self.engine
        base_tag = e._system_tags[SystemTagName.BASE]
        base_tag.set_value("s", e._tick_time)

    def test_base_threshold_uses_base_sec(self):
        program = "1 Mark: A1"
        e = self.engine
        base_tag = e._system_tags[SystemTagName.BASE]
        base_tag.set_value("s", e._tick_time)
        run_engine(e, program, 15)
        self.assertEqual(['A1'], e.interpreter.get_marks())

    def test_base_threshold_uses_base_min(self):
        program = "1 Mark: A1"
        e = self.engine
        base_tag = e._system_tags[SystemTagName.BASE]
        base_tag.set_value("min", e._tick_time)
        run_engine(e, program, 15)
        self.assertEqual([], e.interpreter.get_marks())

    def test_base_threshold_uses_default_base_min(self):
        program = "1 Mark: A1"
        e = self.engine
        run_engine(e, program, 15)
        self.assertEqual([], e.interpreter.get_marks())

    @unittest.skip("Review with Eskild")
    def test_block_time_watch_2(self):
        program = build_program("""
Block: A
    Mark: A1
    Watch: Block Time > 0 sec
        End block
    Mark: X
    Mark: A2
Mark: A3
""")
        # TODO review case with Eskild
        # We'll need look-ahead to avoid X being executed

        i = create_interpreter(program=program)
        run_interpreter(i, max_ticks=30)

        self.assertEqual(["A1", "A3"], i.get_marks())

    def test_threshold_time(self):
        program = """
Base: s
0.0 Mark: a
0.3 Mark: b
0.8 Mark: c
"""
        engine = self.engine
        run_engine(engine, program, 20)
        i = engine.interpreter

        self.assertEqual(["a", "b", "c"], i.get_marks())
        # print_log(i)

        # log_a = next(x for x in i.logs if x.message == 'a')
        # log_b = next(x for x in i.logs if x.message == 'b')
        # log_c = next(x for x in i.logs if x.message == 'c')
        # with self.subTest("a, b"):
        #     self.assert_time_equal(log_a.time + .3, log_b.time)
        # with self.subTest("a, c"):
        #     self.assert_time_equal(log_a.time + .8, log_c.time)

    def test_assert_time_equal(self):
        self.assert_time_equal(1, 1.2, 200)
        self.assert_time_equal(1.8, 1.5, 300)
        self.assertRaises(
            AssertionError,
            lambda: self.assert_time_equal(1.8, 1.3, 400)
        )

    def assert_time_equal(self, time_a: float, time_b: float, milis=200):
        """
        Assert that time difference is less than milis miliseconds.

        Note that the default value depends on the interpreter timer interval.
        """
        diff_milis = round(1000 * abs(time_a - time_b))
        if diff_milis > milis:
            s = f"Difference exceeded {milis} ms.\n" + \
                f"Difference: {diff_milis} ms.\n" + \
                f"time_a:      {time_a:.2f} s.\n" + \
                f"time_b:      {time_b:.2f} s.\n"
            raise AssertionError(s)
        else:
            print(f"Diff {diff_milis} accepted")

    def test_threshold_block_time(self):
        program = """
Base: s
Mark: a
Block: A
    0.3 Mark: b
    End block
Block: B
    0.8 Mark: c
    End block
Mark: d
"""
        engine = self.engine
        run_engine(engine, program, 25)

        i = engine.interpreter

        self.assertEqual(["a", "b", "c", "d"], i.get_marks())

    def test_wait(self):
        program = """
Mark: a
Wait: 0.5 s
Mark: b"""
        engine = self.engine
        run_engine(engine, program, 4)
        i = engine.interpreter
        self.assertEqual(["a"], i.get_marks())

        continue_engine(engine, 4)
        self.assertEqual(["a"], i.get_marks())

        continue_engine(engine, 2)

        self.assertEqual(["a", "b"], i.get_marks())

    @unittest.skip("TODO")
    def test_threshold_column_volume(self):
        raise NotImplementedError()

    @unittest.skip("TODO")
    def test_watch_tag_categorized_value(self):
        program = build_program("""
Watch: LT01 = Full
    Mark: a1
Mark: b
""")
        i = create_interpreter(program=program)
        run_interpreter(i, 5)

    @unittest.skip("TODO")
    def test_change_base_in_program(self):
        # base is also available in scope
        raise NotImplementedError()

    @unittest.skip("TODO")
    def test_change_base_in_scope(self):
        # base is global, a change should remain in place after scope completes
        raise NotImplementedError()


if __name__ == "__main__":
    unittest.main()



import logging
import os
import sentry_sdk
from sentry_sdk.integrations.starlette import StarletteIntegration
from sentry_sdk.integrations.fastapi import FastApiIntegration
from sentry_sdk.integrations.logging import LoggingIntegration

from openpectus import __version__ as openpectus_version, build_number
from openpectus.lang.exec.uod import UnitOperationDefinitionBase

# Sentry docs: https://docs.sentry.io/platforms/python
# Logging docs: https://docs.sentry.io/platforms/python/integrations/logging

# define valid leves for use as sentry event_level filter
_event_levels: list[int] = [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR]
EVENT_LEVEL_DEFAULT: str = logging.getLevelName(logging.WARNING)
EVENT_LEVEL_NAMES: list[str] = [logging.getLevelName(level) for level in _event_levels]

SENTRY_DEBUG = False


def _get_event_level_from_name(event_level_name: str) -> int:
    event_level_name = event_level_name.upper()
    if event_level_name not in EVENT_LEVEL_NAMES:
        raise ValueError(f"Invalid event_level_name specified: {event_level_name}. " +
                         f"Must be one of {','.join(EVENT_LEVEL_NAMES)}")
    return logging.getLevelNamesMapping()[event_level_name]


def init_aggregator(event_level_name: str):
    """ Initialize sentry logging for the aggregator component.

    Note: call after initializing logging.

    - Enables integrations FastAPI and Starlette
    - Attaches component='aggregator' to all events
    """
    if os.getenv('SENTRY_DSN'):
        print("Sentry is active")
    else:
        print("Sentry is not active")

    event_level_int = _get_event_level_from_name(event_level_name)

    sentry_sdk.init(
        default_integrations=False,
        auto_enabling_integrations=False,
        integrations=[
            StarletteIntegration(transaction_style="endpoint"),
            FastApiIntegration(transaction_style="endpoint"),
            LoggingIntegration(
                level=logging.INFO,        # Capture info and above as breadcrumbs
                event_level=event_level_int   # Send records as events
            )
        ],
        auto_session_tracking=False,
        send_default_pii=False,
        release=openpectus_version,
        # Set traces_sample_rate to 1.0 to capture 100%
        # of transactions for performance monitoring.
        # traces_sample_rate=1.0,
        # enable_tracing=True,
        # Set profiles_sample_rate to 1.0 to profile 100%
        # of sampled transactions.
        # We recommend adjusting this value in production.
        # profiles_sample_rate=1.0,
        debug=SENTRY_DEBUG,
        # server_name is set to machine name by default, which is fine
        # server_name="aggregator",
        # before_send works but we'd rather use tags instead
        # before_send=agg_before_send,
        # before_send_transaction is not currently called
        # maybe transactions are for traces only?
        # before_send_transaction=agg_before_send_transaction,
    )
    # this seems to attach component to all events
    sentry_sdk.set_tag("component", "aggregator")
    sentry_sdk.set_tag("build_number", build_number)

    with sentry_sdk.new_scope():
        sentry_sdk.capture_message("Aggregator sentry logging initialized")

def init_engine(event_level_name: str):
    """ Initialize sentry logging for the engine component.

    - Attaches tag component='engine' to all events
    """

    event_level_int = _get_event_level_from_name(event_level_name)

    sentry_sdk.init(
        # default_integrations=False,
        auto_enabling_integrations=False,
        integrations=[
            LoggingIntegration(
                level=logging.INFO,        # Capture info and above as breadcrumbs
                event_level=event_level_int   # Send records as events
            )
        ],
        auto_session_tracking=False,
        send_default_pii=False,
        release=openpectus_version,
        debug=SENTRY_DEBUG,
    )
    # this seems to attach component to all events
    sentry_sdk.set_tag("component", "engine")
    sentry_sdk.set_tag("build_number", build_number)

    with sentry_sdk.new_scope():
        sentry_sdk.capture_message("Engine sentry logging initialized")

def set_engine_uod(uod: UnitOperationDefinitionBase):
    """ Attaches tags uod_instrument and uod_location to events. """
    sentry_sdk.set_tag("uod_instrument", uod.instrument)
    sentry_sdk.set_tag("uod_location", uod.location)

def engine_method_set(pcode: str):
    """ Attaches tag 'method' with pcode source to events """
    # Note: There is no great format to use but this works reasonably well.
    # To recover pcode formatting, replace \\n with \n (extended/regex) in vscode or npp
    dct = {"pcode": pcode}
    sentry_sdk.set_tag("method", dct)



import hashlib
import json
import os
import unittest

from openpectus.aggregator.aggregator_server import AggregatorServer
from fastapi.testclient import TestClient

project_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
fastapi = AggregatorServer().fastapi
client = TestClient(fastapi)


class AggregatorOpenAPIApiTest(unittest.TestCase):
    def test_read_openapi_docs(self):
        response = client.get("/docs")
        self.assertEqual(200, response.status_code)

    def test_read_openapi_spec(self):
        response = client.get("/openapi.json")
        self.assertEqual(200, response.status_code)
        spec = response.json()
        self.assertEqual(fastapi.title, spec["info"]["title"])

    def test_write_openapi_spec_to_file_and_compare_with_existing(self):
        response = client.get("/openapi.json")
        self.assertEqual(200, response.status_code)
        openapi_file = os.path.join(project_path, "frontend", "openapi.json")
        current_md5, updated_md5 = "", ""
        with open(openapi_file, "rb") as f:
            current_md5 = hashlib.md5(f.read()).hexdigest()

        # parsed and pretty printed
        with open(openapi_file, "wt") as f:
            json.dump(response.json(), f, indent=2)

        with open(openapi_file, "rb") as f:
            updated_md5 = hashlib.md5(f.read()).hexdigest()

        self.assertEqual(current_md5, updated_md5, f"""
The generated Aggregator API specification does not match the one stored in:
{openapi_file}

You probably forgot to update it using the update script:
(pectus) aggregator> python generate_openapi_spec_and_typescript_interfaces.py
""")


if __name__ == "__main__":
    unittest.main()



"""Bootstrap

Revision ID: 646d56d49b76
Revises: 
Create Date: 2024-01-29 15:08:39.310025

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '646d56d49b76'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('PlotLogs',
    sa.Column('engine_id', sa.String(), nullable=False),
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_PlotLogs'))
    )
    op.create_table('ProcessUnits',
    sa.Column('engine_id', sa.String(), nullable=False),
    sa.Column('state', sa.String(), nullable=False),
    sa.Column('location', sa.String(), nullable=False),
    sa.Column('runtime_msec', sa.Integer(), nullable=False),
    sa.Column('current_user_role', sa.String(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_ProcessUnits'))
    )
    op.create_table('RecentRunMethodAndStates',
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('method', sa.JSON(), nullable=False),
    sa.Column('state', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRunMethodAndStates'))
    )
    op.create_table('RecentRunPlotConfigurations',
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('plot_configuration', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRunPlotConfigurations'))
    )
    op.create_table('RecentRunRunLogs',
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('run_log', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRunRunLogs'))
    )
    op.create_table('RecentRuns',
    sa.Column('engine_id', sa.String(), nullable=False),
    sa.Column('run_id', sa.String(), nullable=False),
    sa.Column('computer_name', sa.String(), nullable=False),
    sa.Column('uod_name', sa.String(), nullable=False),
    sa.Column('started_date', sa.DateTime(), nullable=False),
    sa.Column('completed_date', sa.DateTime(), nullable=False),
    sa.Column('contributors', sa.JSON(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_RecentRuns'))
    )
    op.create_table('PlotLogEntries',
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('value_unit', sa.String(), nullable=True),
    sa.Column('value_type', sa.Enum('STRING', 'FLOAT', 'INT', 'CHOICE', 'NONE', name='processvaluetype'), nullable=False),
    sa.Column('plot_log_id', sa.Integer(), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['plot_log_id'], ['PlotLogs.id'], name=op.f('fk_PlotLogEntries_plot_log_id_PlotLogs')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_PlotLogEntries'))
    )
    op.create_table('PlotLogEntryValues',
    sa.Column('plot_log_entry_id', sa.Integer(), nullable=False),
    sa.Column('tick_time', sa.Float(), nullable=False),
    sa.Column('value_str', sa.String(), nullable=True),
    sa.Column('value_float', sa.Float(), nullable=True),
    sa.Column('value_int', sa.Integer(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['plot_log_entry_id'], ['PlotLogEntries.id'], name=op.f('fk_PlotLogEntryValues_plot_log_entry_id_PlotLogEntries')),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_PlotLogEntryValues'))
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('PlotLogEntryValues')
    op.drop_table('PlotLogEntries')
    op.drop_table('RecentRuns')
    op.drop_table('RecentRunRunLogs')
    op.drop_table('RecentRunPlotConfigurations')
    op.drop_table('RecentRunMethodAndStates')
    op.drop_table('ProcessUnits')
    op.drop_table('PlotLogs')
    # ### end Alembic commands ###
